with Ada.Text_IO; use Ada.Text_IO;
with Name_Table;
with Libraries;

package body Flags is
   function Option_Warning (Opt: String) return Boolean is
   begin
      if Opt = "undriven" then
         Warn_Undriven := True;
      elsif Opt = "binding" then
         Warn_Default_Binding := True;
      elsif Opt = "reserved" then
         Warn_Reserved_Word := True;
      else
         return False;
      end if;
      return True;
   end Option_Warning;

   function Parse_Option (Opt: String) return Boolean is
      Beg: Integer := Opt'First;
   begin
      if Opt = "-C" or else Opt = "--mb-comments" then
         Mb_Comment := True;
      elsif Opt = "--bootstrap" then
         Bootstrap := True;
      elsif Opt = "--syn-binding" then
         Flag_Syn_Binding := True;
      elsif Opt'Length > 7 and then Opt (Beg .. Beg + 6) = "--warn-" then
         return Option_Warning (Opt (Beg + 7 .. Opt'Last));
      elsif Opt'Length > 7 and then Opt (Beg .. Beg + 6) = "--work=" then
         Libraries.Work_Library_Name :=
           Name_Table.Get_Identifier (Opt (Beg + 7 .. Opt'Last));
      elsif Opt = "--no-vital-checks" then
         Flag_Vital_Checks := False;
      elsif Opt = "--vital-checks" then
         Flag_Vital_Checks := True;
      elsif Opt = "-dp" then
         Dump_Parse := True;
      elsif Opt = "-ds" then
         Dump_Sem := True;
      elsif Opt = "-dc" then
         Dump_Canon := True;
      elsif Opt = "-da" then
         Dump_Annotate := True;
      elsif Opt = "--dall" then
         Dump_All := True;
      elsif Opt'Length > 5 and then Opt (Beg .. Beg + 5) = "--std=" then
         if Opt'Length = 8 then
            if Opt (Beg + 6 .. Beg + 7) = "87" then
               Vhdl_Std := Vhdl_87;
            elsif Opt (Beg + 6 .. Beg + 7) = "93" then
               Vhdl_Std := Vhdl_93;
            elsif Opt (Beg + 6 .. Beg + 7) = "00" then
               Vhdl_Std := Vhdl_00;
            elsif Opt (Beg + 6 .. Beg + 7) = "02" then
               Vhdl_Std := Vhdl_02;
            else
               return False;
            end if;
         elsif Opt'Length = 9 and then Opt (Beg + 6 .. Beg + 8) = "93c" then
            Vhdl_Std := Vhdl_93c;
         else
            return False;
         end if;
      elsif Opt'Length > 2 and then Opt (Beg .. Beg + 1) = "-P" then
         Libraries.Add_Library_Path (Opt (Beg + 2 .. Opt'Last));
      elsif Opt'Length > 10 and then Opt (Beg .. Beg + 9) = "--workdir=" then
         Libraries.Set_Work_Library_Path (Opt (Beg + 10 .. Opt'Last));
      elsif Opt = "--lall" then
         List_All := True;
      elsif Opt = "-lv" then
         List_Verbose := True;
      elsif Opt = "-ls" then
         List_Sem := True;
      elsif Opt = "-lc" then
         List_Canon := True;
      elsif Opt = "-la" then
         List_Annotate := True;
      elsif Opt = "-v" then
         Verbose := True;
      elsif Opt'Length > 17
        and then Opt (Beg .. Beg + 17) = "--time-resolution="
      then
         Beg := Beg + 18;
         if Opt (Beg .. Beg + 1) = "fs" then
            Time_Resolution := 'f';
         elsif Opt (Beg .. Beg + 1) = "ps" then
            Time_Resolution := 'p';
         elsif Opt (Beg .. Beg + 1) = "ns" then
            Time_Resolution := 'n';
         elsif Opt (Beg .. Beg + 1) = "us" then
            Time_Resolution := 'u';
         elsif Opt (Beg .. Beg + 1) = "ms" then
            Time_Resolution := 'm';
         elsif Opt (Beg .. Beg + 2) = "sec" then
            Time_Resolution := 's';
         elsif Opt (Beg .. Beg + 2) = "min" then
            Time_Resolution := 'M';
         elsif Opt (Beg .. Beg + 1) = "hr" then
            Time_Resolution := 'h';
         else
            return False;
         end if;
      else
         return False;
      end if;
      return True;
   end Parse_Option;

   -- Disp help about these options.
   procedure Disp_Options_Help is
   begin
      Put_Line ("Main options:");
      Put_Line ("  --work=LIB  use LIB as work library");
      Put_Line ("  --workdir=DIR    use DIR for the file library");
      Put_Line ("  -PPATH      add PATH in the library path list");
      Put_Line ("  --std=87    select vhdl 87 standard");
      Put_Line ("  --std=93    select vhdl 93 standard");
      Put_Line ("  --std=93c   select vhdl 93 standard and allow 87 syntax");
      Put_Line ("  --[no-]vital-checks  do [not] check VITAL restrictions");
      Put_Line ("Warnings:");
      Put_Line ("  --warn-undriven  disp undriven signals");
      Put_Line ("  --warn-binding   disp component not bound by default");
      Put_Line ("  --warn-reserved  warns use of 93 reserved words in vhdl87");
--    Put_Line ("Simulation option:");
--    Put_Line ("  --time-resolution=UNIT   set the resolution of type time");
--    Put_Line ("            UNIT can be fs, ps, ns, us, ms, sec, min or hr");
--    Put_Line ("  --assert-level=LEVEL     set the level which stop the");
--    Put_Line ("           simulation.  LEVEL is note, warning, error,");
--    Put_Line ("           failure or none");
      Put_Line ("Illegal extensions:");
      Put_Line ("  -C  --mb-comments  allow multi-bytes chars in a comment");
      Put_Line ("  --bootstrap        allow --work=std");
      Put_Line ("  --syn-binding      use synthesis default binding rule");
      Put_Line ("Compilation dump:");
      Put_Line ("  -dp         tree after parsing");
      Put_Line ("  -ds         tree after semantics");
      Put_Line ("  -da         tree after annotate");
      Put_Line ("  --dall      -dX options apply to all files");
      Put_Line ("Compilation list:");
      Put_Line ("  -ls         after semantics");
      Put_Line ("  -lc         after canon");
      Put_Line ("  -la         after annotation");
      Put_Line ("  --lall      -lX options apply to all files");
      Put_Line ("  -lv         verbose list");
      Put_Line ("  -v          disp compilation stages");
   end Disp_Options_Help;

end Flags;
