with Ada.Text_IO; use Ada.Text_IO;
with Flags;
with Types;
with Name_Table;
with Sem;
with Errorout; use Errorout;
with Canon;

package body Back_End is
   procedure Finish_Compilation
     (Unit : Iir_Design_Unit; Main : Boolean := False)
   is
      pragma Unreferenced (Main);
   begin
      if Flags.Verbose then
         Put_Line ("semantize " & Disp_Node (Get_Library_Unit (Unit)));
      end if;

      Sem.Semantic (Unit);

      if Errorout.Nbr_Errors > 0 then
         raise Compilation_Error;
      end if;

      if Flags.Flag_Elaborate then
         Canon.Canonicalize (Unit);

         declare
            Config : Iir_Design_Unit;
            Lib : Iir;
         begin
            Lib := Get_Library_Unit (Unit);
            if Get_Kind (Lib) = Iir_Kind_Architecture_Declaration then
               Config := Canon.Create_Default_Configuration_Declaration (Lib);
               Set_Default_Configuration_Declaration (Lib, Config);
            end if;
         end;
      end if;
   end Finish_Compilation;

   procedure Internal_Compilation is
   begin
      raise Program_Error;
   end Internal_Compilation;

   -- Transform a library identifier into a file name.
   -- Very simple mechanism: just add '-simVV.cf' extension, where VV
   -- is the version.
   function Library_To_File_Name (Library: Iir_Library_Declaration)
     return String
   is
      use Types;

      Id : Name_Id;
   begin
      Id := Get_Identifier (Library);
      case Flags.Vhdl_Std is
         when Vhdl_87 =>
            return Name_Table.Image (Id) & "-obj87.cf";
         when Vhdl_93c
           | Vhdl_93
           | Vhdl_00
           | Vhdl_02 =>
            return Name_Table.Image (Id) & "-obj93.cf";
      end case;
   end Library_To_File_Name;
end Back_End;
