with Grt.Types; use Grt.Types;
with Grt.Signals; use Grt.Signals;
with Grt.Processes;
--with Grt.Hierarchy; use Grt.Hierarchy;
with Ada.Characters.Latin_1;
with System; use System;
--with Grt.Errors; use Grt.Errors;

package body Grt.Disp is
   procedure Put (Stream : FILEs; Str : String)
   is
      S : size_t;
   begin
      S := fwrite (Str'Address, Str'Length, 1, Stream);
   end Put;

   procedure Put (Str : String)
   is
      S : size_t;
   begin
      S := fwrite (Str'Address, Str'Length, 1, stdout);
   end Put;

   procedure Put (C : Character)
   is
      R : int;
   begin
      R := fputc (Character'Pos (C), stdout);
   end Put;

   procedure Put (Str : Ghdl_C_String)
   is
      Len : Natural;
      S : size_t;
   begin
      Len := strlen (Str);
      S := fwrite (Str (1)'Address, size_t (Len), 1, stdout);
   end Put;


   procedure New_Line is
   begin
      Put (Ada.Characters.Latin_1.LF);
   end New_Line;

   procedure Put_Line (Str : String)
   is
   begin
      Put (Str);
      New_Line;
   end Put_Line;

--    function Get_String (Str : Ghdl_Str_Len_Type) return String
--    is
--    begin
--       if Str.Str = null then
--          return ''' & Character'Val (Str.Len) & ''';
--       else
--          return Str.Str (1 .. Str.Len);
--       end if;
--    end Get_String;

   procedure Put_Str_Len (Stream : FILEs; Str : Ghdl_Str_Len_Type)
   is
      S : String (1 .. 3);
   begin
      if Str.Str = null then
         S (1) := ''';
         S (2) := Character'Val (Str.Len);
         S (3) := ''';
         Put (Stream, S);
      else
         Put (Stream, Str.Str (1 .. Str.Len));
      end if;
   end Put_Str_Len;

--    procedure Put_Trim (Stream : FILEs; Str : String)
--    is
--       Start : Natural;
--    begin
--       Start := Str'First;
--       while Start <= Str'Last and then Str (Start) = ' ' loop
--          Start := Start + 1;
--       end loop;
--       Put (Stream, Str (Start .. Str'Last));
--    end Put_Trim;

   generic
      type Ntype is range <>;
      Max_Len : Natural;
   procedure Put_Ntype (Stream : FILEs; N : Ntype);

   procedure Put_Ntype (Stream : FILEs; N : Ntype)
   is
      Str : String (1 .. Max_Len);
      P : Natural := Str'Last;
      V : Ntype;
   begin
      if N > 0 then
         V := -N;
      else
         V := N;
      end if;
      loop
         Str (P) := Character'Val (48 - (V rem 10));
         V := V / 10;
         exit when V = 0;
         P := P - 1;
      end loop;
      if N < 0 then
         P := P - 1;
         Str (P) := '-';
      end if;
      Put (Stream, Str (P .. Max_Len));
   end Put_Ntype;

   procedure Put_I32_1 is new Put_Ntype (Ntype => Ghdl_I32, Max_Len => 11);

   procedure Put_I32 (Stream : FILEs; I32 : Ghdl_I32) renames Put_I32_1;

   procedure Put_I64_1 is new Put_Ntype (Ntype => Ghdl_I64, Max_Len => 20);
   procedure Put_I64 (Stream : FILEs; I64 : Ghdl_I64) renames Put_I64_1;

   procedure Put_F64 (Stream : FILEs; F64 : Ghdl_F64)
   is
      procedure fprintf (Stream : FILEs;
                         Template : System.Address;
                         Arg : Ghdl_F64);
      pragma Import (C, fprintf);

      Str : constant String := "%g" & Character'Val (0);
   begin
      fprintf (Stream, Str'Address, F64);
   end Put_F64;

   Hex_Map : constant array (0 .. 15) of Character := "0123456789ABCDEF";

   procedure Put (Stream : FILEs; Addr : System.Address)
   is
      Res : String (1 .. System.Word_Size / 4);
      Val : Integer_Address := To_Integer (Addr);
   begin
      for I in reverse Res'Range loop
         Res (I) := Hex_Map (Natural (Val and 15));
         Val := Val / 16;
      end loop;
      Put (Stream, Res);
   end Put;

--   procedure Put_E8 (Stream : FILEs; E8 : Ghdl_E8; Type_Desc : Ghdl_Desc_Ptr)
--    is
--    begin
--       Put_Str_Len (Stream, Type_Desc.E8.Values (Natural (E8)));
--    end Put_E8;

   --procedure Put_E32
   --  (Stream : FILEs; E32 : Ghdl_E32; Type_Desc : Ghdl_Desc_Ptr)
   --is
   --begin
   --   Put_Str_Len (Stream, Type_Desc.E32.Values (Natural (E32)));
   --end Put_E32;

   procedure Put_Dir (Stream : FILEs; Dir : Ghdl_Dir_Type) is
   begin
      case Dir is
         when Dir_To =>
            Put (Stream, " to ");
         when Dir_Downto =>
            Put (Stream, " downto ");
      end case;
   end Put_Dir;

   procedure Put_Sig_Index (Sig : Sig_Table_Index)
   is
   begin
      Put_I32 (stdout, Ghdl_I32 (Sig));
   end Put_Sig_Index;

   procedure Put_Sig_Range (Sig : Sig_Table_Range)
   is
   begin
      Put_Sig_Index (Sig.First);
      if Sig.Last /= Sig.First then
         Put ("-");
         Put_Sig_Index (Sig.Last);
      end if;
   end Put_Sig_Range;

   procedure Disp_Time (T : Std_Time) is
   begin
      if T = Std_Time'First then
         Put ("-Inf");
      else
         Put_I64 (stdout, Ghdl_I64 (T));
      end if;
   end Disp_Time;

   procedure Disp_Now
   is
      use Grt.Processes;
   begin
      Put ("Now is ");
      Disp_Time (Current_Time);
      Put (" +");
      Put_I32 (stdout, Ghdl_I32 (Current_Delta));
      New_Line;
   end Disp_Now;

   procedure Disp_Signals_Order is
   begin
      for I in Propagation.First .. Propagation.Last loop
         case Propagation.Table (I).Kind is
            when Drv_One_Driver =>
               Put ("Drv (drv) ");
            when Eff_One_Driver =>
               Put ("Eff (drv) ");
            when Drv_One_Port =>
               Put ("Drv (prt) ");
            when Eff_One_Port =>
               Put ("Eff (prt) ");
            when Imp_Guard =>
               Put ("Guard ");
            when Imp_Stable =>
               Put ("Stable ");
            when Imp_Quiet =>
               Put ("imp quiet");
            when Imp_Transaction =>
               Put ("imp transaction");
            when Imp_Delayed =>
               Put ("imp delayed");
            when Eff_Actual =>
               Put ("Actual ");
            when Eff_Multiple =>
               Put ("Eff resolved ");
            when Drv_One_Resolved =>
               Put ("Drv 1 resolved ");
            when Eff_One_Resolved =>
               Put ("Eff 1 resolved ");
            when In_Conversion =>
               Put ("In conv ");
            when Out_Conversion =>
               Put ("Out conv ");
            when Drv_Error =>
               Put ("Drv error");
            when Drv_Multiple =>
               Put ("drv multiple");
         end case;
         case Propagation.Table (I).Kind is
            when Drv_One_Driver
              | Eff_One_Driver
              | Drv_One_Port
              | Eff_One_Port
              | Drv_One_Resolved
              | Eff_One_Resolved
              | Imp_Guard
              | Imp_Stable
              | Eff_Actual =>
               Put_Sig_Index (Signal_Ptr_To_Index (Propagation.Table (I).Sig));
               New_Line;
            when Eff_Multiple =>
               Put_Sig_Range (Propagation.Table (I).Resolv.Sig_Range);
               New_Line;
            when In_Conversion
              | Out_Conversion =>
               declare
                  Conv : Sig_Conversion_Acc;
               begin
                  Conv := Propagation.Table (I).Conv;
                  Put_Sig_Range (Conv.Src);
                  Put (" -> ");
                  Put_Sig_Range (Conv.Dest);
                  New_Line;
               end;
            when others =>
               null;
         end case;
      end loop;
   end Disp_Signals_Order;
end Grt.Disp;
