with Grt.Errors;
with Grt.Vcd;
with Grt.Vpi;
with Grt.Stacks;
with Grt.Processes;
with Grt.Signals;
with Grt.Options;
with Ada.Unchecked_Conversion;
with Grt.Disp_Rti;
with Grt.Hooks;

--  The following packages are not referenced in this package.
--  They are with'ed in order to be present in the binary.
pragma Warnings (Off);
with Grt.Files;
with Grt.Types;
with Grt.Lib;
with Grt.Shadow_Ieee;
with Grt.Images;
with Grt.Names;
with Grt.Vital_Annotate;
pragma Warnings (On);

package body Grt.Main is
   procedure Ghdl_Elaborate;
   pragma Import (C, Ghdl_Elaborate, "__ghdl_ELABORATE");

   procedure Run
   is
      use Grt.Errors;
      Stop : Boolean;
   begin
      Grt.Vcd.Register;
      Grt.Vpi.Register;

      Grt.Options.Decode (Stop);

      if not Stop then
         Grt.Stacks.Stack_Init;

         Grt.Hooks.Call_Init_Hooks;

         Grt.Processes.Init;

         Ghdl_Elaborate;

         if Grt.Options.Flag_Dump_Rti then
            Grt.Disp_Rti.Disp_All;
         end if;

         Grt.Signals.Order_All_Signals;

         Grt.Processes.Start;

         if Expect_Failure then
            Expect_Failure := False;
            Error ("error expected, but none occured");
         end if;
      end if;
   end Run;

   function Ghdl_Main (Argc : Integer; Argv : System.Address)
      return Integer
   is
      function To_Argv_Type is new Ada.Unchecked_Conversion
        (Source => System.Address, Target => Grt.Options.Argv_Type);
   begin
      Grt.Options.Argc := Argc;
      Grt.Options.Argv := To_Argv_Type (Argv);

      Grt_Init;
      Run;
      return 0;
   end Ghdl_Main;
end Grt.Main;
