with Sem_Scopes; use Sem_Scopes;
with Sem_Names; use Sem_Names;
with Errorout; use Errorout;
with Types; use Types;
with Sem_Expr; use Sem_Expr;
with Std_Names;
with Tokens;
with Sem_Specs; use Sem_Specs;
with Flags;
with Std_Package; use Std_Package;
with Iir_Chains;
with Evaluation; use Evaluation;
with Name_Table;
with Iirs_Utils; use Iirs_Utils;
with Sem_Types; use Sem_Types;
with Sem; use Sem;
with Iir_Chains; use Iir_Chains;

package body Sem_Decls is
   procedure Sem_Interface_Chain (Interface_Chain: Iir;
                                  Interface_Kind : Interface_Kind_Type)
   is
      El, A_Type: Iir;
      Proxy : Iir_Proxy;
      Default_Value: Iir;
   begin
      if Interface_Chain = Null_Iir then
         return;
      end if;
      El := Interface_Chain;
      while El /= Null_Iir loop
         --  Avoid the reanalysed duplicated types.
         --  This is not an optimization, since the unanalysed type must have
         --  been freed.
         A_Type := Get_Type (El);
         if Get_Kind (A_Type) = Iir_Kind_Proxy then
            Proxy := A_Type;
            A_Type := Get_Type (Get_Proxy (Proxy));
            Default_Value := Get_Default_Value (Get_Proxy (Proxy));
            Free_Iir (Proxy);
         else
            A_Type := Sem_Subtype_Indication (A_Type);
            Default_Value := Get_Default_Value (El);
            if Default_Value /= Null_Iir and then A_Type /= Null_Iir then
               Deferred_Constant_Allowed := True;
               Default_Value := Sem_Expression (Default_Value, A_Type);
               Deferred_Constant_Allowed := False;
               Check_Read (Default_Value);
            end if;
         end if;

         Set_Base_Name (El, El);
         Set_Name_Staticness (El, Locally);

         if A_Type /= Null_Iir then
            Set_Type (El, A_Type);

            if Get_Kind (El) = Iir_Kind_Signal_Interface_Declaration then
               case Get_Signal_Kind (El) is
                  when Iir_No_Signal_Kind =>
                     null;
                  when Iir_Bus_Kind =>
                     --  FIXME: where this test comes from ?
                     if False
                       and
                       (Get_Kind (A_Type) not in Iir_Kinds_Subtype_Definition
                        or else Get_Resolution_Function (A_Type) = Null_Iir)
                     then
                        Error_Msg_Sem
                          (Disp_Node (A_Type)
                           & " of guarded " & Disp_Node (El)
                           & " is not resolved", El);
                     end if;
                  when Iir_Register_Kind =>
                     Error_Msg_Sem
                       ("interface signal can't be of kind register", El);
               end case;
            end if;

            --  LRM93 4.3.2
            --  For an interface constant declaration or an interface signal
            --  declaration, the subtype indication must define a subtype
            --  that is neither a file type nor an access type.
            case Get_Kind (El) is
               when Iir_Kind_Constant_Interface_Declaration
                 | Iir_Kind_Signal_Interface_Declaration =>
                  case Get_Kind (Get_Base_Type (A_Type)) is
                     when Iir_Kind_File_Type_Definition =>
                        Error_Msg_Sem
                          ("constant or signal formal type can't "
                           & "be a file type", El);
                     when Iir_Kind_Access_Type_Definition =>
                        Error_Msg_Sem
                          ("constant or signal formal type can't "
                           & "be an access type", El);
                     when others =>
                        null;
                  end case;
               when Iir_Kind_Variable_Interface_Declaration =>
                  if Get_Kind (Get_Base_Type (A_Type))
                    = Iir_Kind_File_Type_Definition
                    and then Flags.Vhdl_Std >= Vhdl_93
                  then
                     Error_Msg_Sem
                       ("variable formal type can't be a file type (vhdl 93)",
                        El);
                  end if;
               when Iir_Kind_File_Interface_Declaration =>
                  if Get_Kind (Get_Base_Type (A_Type))
                    /= Iir_Kind_File_Type_Definition
                  then
                     Error_Msg_Sem
                       ("file formal type must be a file type", El);
                  end if;
               when others =>
                  --  El is not an interface.
                  raise Internal_Error;
            end case;

            if Default_Value /= Null_Iir then
               Set_Default_Value (El, Default_Value);
            end if;
         else
            Set_Type (El, Error_Type);
         end if;

         Sem_Scopes.Add_Name (El);

         --  By default, interface are not static.
         --  This may be changed just below.
         Set_Expr_Staticness (El, None);

         case Interface_Kind is
            when Interface_Generic =>
               --  LRM93 1.1.1
               --  The generic list in the formal generic clause defines
               --  generic constants whose values may be determined by the
               --  environment.
               if Get_Kind (El) /= Iir_Kind_Constant_Interface_Declaration then
                  Error_Msg_Sem
                    ("generic " & Disp_Node (El) & " must be a constant",
                     El);
               else
                  --   LRM93 7.4.2 (Globally static primaries)
                  --   3. a generic constant.
                  Set_Expr_Staticness (El, Globally);
               end if;
            when Interface_Port =>
               if Get_Kind (El) /= Iir_Kind_Signal_Interface_Declaration then
                  Error_Msg_Sem
                    ("port " & Disp_Node (El) & " must be a signal", El);
               end if;
            when Interface_Procedure
              | Interface_Function =>
               if Get_Kind (El) = Iir_Kind_Variable_Interface_Declaration
                 and then Interface_Kind = Interface_Function
               then
                  Error_Msg_Sem ("variable interface parameter are not "
                                 & "allowed for a function (use a constant)",
                                 El);
               end if;
               case Get_Mode (El) is
                  when Iir_Unknown_Mode =>
                     raise Internal_Error;
                  when Iir_In_Mode =>
                     null;
                  when Iir_Inout_Mode
                    | Iir_Out_Mode =>
                     if Interface_Kind = Interface_Function
                       and then
                       Get_Kind (El) /= Iir_Kind_File_Interface_Declaration
                     then
                        Error_Msg_Sem ("inout or out is not an allowed mode " &
                                       "for a function parameter", El);
                     end if;
                  when Iir_Buffer_Mode
                    | Iir_Linkage_Mode =>
                     Error_Msg_Sem ("buffer or linkage is not an allowed mode "
                                    & "for a subprogram parameter", El);
               end case;
         end case;
         El := Get_Chain (El);
      end loop;

      --  Make all name visible.
      El := Interface_Chain;
      while El /= Null_Iir loop
         Name_Visible (El);
         El := Get_Chain (El);
      end loop;
   end Sem_Interface_Chain;

   --  LRM93 7.2.2
   --  A discrete array is a one-dimensional array whose elements are of a
   --  discrete type.
   function Is_Discrete_Array (Def : Iir) return Boolean
   is
   begin
      case Get_Kind (Def) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Array_Subtype_Definition =>
            null;
         when others =>
            raise Internal_Error;
            -- return False;
      end case;
      if Get_Nbr_Elements (Get_Index_Subtype_List (Def)) /= 1 then
         return False;
      end if;
      if Get_Kind (Get_Element_Subtype (Def))
        not in Iir_Kinds_Discrete_Type_Definition
      then
         return False;
      end if;
      return True;
   end Is_Discrete_Array;

   procedure Create_Implicit_File_Primitives
     (Decl : Iir_Type_Declaration; Type_Definition : Iir_File_Type_Definition)
   is
      use Iir_Chains.Interface_Declaration_Chain_Handling;
      Type_Mark: Iir;
      Proc: Iir_Implicit_Procedure_Declaration;
      Func: Iir_Implicit_Function_Declaration;
      Interface: Iir;
      Loc : Location_Type;
      File_Interface_Kind : Iir_Kind;
      Last_Interface : Iir;
      Last : Iir;
   begin
      Last := Decl;
      Type_Mark := Get_Type_Mark (Type_Definition);
      Loc := Get_Location (Decl);

      if Flags.Vhdl_Std >= Vhdl_93c then
         for I in 1 .. 2 loop
            --  Create the implicit file_open (form 1) declaration.
            --  Create the implicit file_open (form 2) declaration.
            Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
            Set_Location (Proc, Loc);
            Set_Parent (Proc, Get_Parent (Decl));
            Set_Identifier (Proc, Std_Names.Name_File_Open);
            Set_Type_Reference (Proc, Decl);
            Build_Init (Last_Interface);
            case I is
               when 1 =>
                  Set_Implicit_Definition (Proc, Iir_Predefined_File_Open);
               when 2 =>
                  Set_Implicit_Definition (Proc,
                                           Iir_Predefined_File_Open_Status);
                  --  status : out file_open_status.
                  Interface :=
                    Create_Iir (Iir_Kind_Variable_Interface_Declaration);
                  Set_Location (Interface, Loc);
                  Set_Identifier (Interface, Std_Names.Name_Status);
                  Set_Type (Interface,
                            Std_Package.File_Open_Status_Type_Definition);
                  Set_Mode (Interface, Iir_Out_Mode);
                  Append (Last_Interface, Proc, Interface);
            end case;
            --  File F : FT
            Interface := Create_Iir (Iir_Kind_File_Interface_Declaration);
            Set_Location (Interface, Loc);
            Set_Identifier (Interface, Std_Names.Name_F);
            Set_Type (Interface, Type_Definition);
            Set_Mode (Interface, Iir_Inout_Mode);
            Append (Last_Interface, Proc, Interface);
            --  External_Name : in STRING
            Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
            Set_Location (Interface, Loc);
            Set_Identifier (Interface, Std_Names.Name_External_Name);
            Set_Type (Interface, Std_Package.String_Type_Definition);
            Set_Mode (Interface, Iir_In_Mode);
            Append (Last_Interface, Proc, Interface);
            --  Open_Kind : in File_Open_Kind := Read_Mode.
            Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
            Set_Location (Interface, Loc);
            Set_Identifier (Interface, Std_Names.Name_Open_Kind);
            Set_Type (Interface, Std_Package.File_Open_Kind_Type_Definition);
            Set_Mode (Interface, Iir_In_Mode);
            Set_Default_Value (Interface,
                               Std_Package.File_Open_Kind_Read_Mode);
            Append (Last_Interface, Proc, Interface);
            -- Add it to the list.
            Insert_Incr (Last, Proc);
         end loop;

         --  Create the implicit file_close declaration.
         Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
         Set_Identifier (Proc, Std_Names.Name_File_Close);
         Set_Location (Proc, Loc);
         Set_Parent (Proc, Get_Parent (Decl));
         Set_Implicit_Definition (Proc, Iir_Predefined_File_Close);
         Set_Type_Reference (Proc, Decl);
         Build_Init (Last_Interface);
         Interface := Create_Iir (Iir_Kind_File_Interface_Declaration);
         Set_Identifier (Interface, Std_Names.Name_F);
         Set_Location (Interface, Loc);
         Set_Type (Interface, Type_Definition);
         Set_Mode (Interface, Iir_Inout_Mode);
         Append (Last_Interface, Proc, Interface);
         -- Add it to the list.
         Insert_Incr (Last, Proc);
      end if;

      if Flags.Vhdl_Std = Vhdl_87 then
         File_Interface_Kind := Iir_Kind_Variable_Interface_Declaration;
      else
         File_Interface_Kind := Iir_Kind_File_Interface_Declaration;
      end if;

      -- Create the implicit procedure read declaration.
      Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
      Set_Identifier (Proc, Std_Names.Name_Read);
      Set_Location (Proc, Loc);
      Set_Parent (Proc, Get_Parent (Decl));
      Set_Type_Reference (Proc, Decl);
      Build_Init (Last_Interface);
      Interface := Create_Iir (File_Interface_Kind);
      Set_Identifier (Interface, Std_Names.Name_F);
      Set_Location (Interface, Loc);
      Set_Type (Interface, Type_Definition);
      Set_Mode (Interface, Iir_Inout_Mode);
      Append (Last_Interface, Proc, Interface);
      Interface := Create_Iir (Iir_Kind_Variable_Interface_Declaration);
      Set_Identifier (Interface, Std_Names.Name_Value);
      Set_Location (Interface, Loc);
      Set_Type (Interface, Type_Mark);
      Set_Mode (Interface, Iir_Out_Mode);
      Append (Last_Interface, Proc, Interface);
      case Get_Kind (Type_Mark) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
            Interface := Create_Iir (Iir_Kind_Variable_Interface_Declaration);
            Set_Identifier (Interface, Std_Names.Name_Length);
            Set_Location (Interface, Loc);
            Set_Type (Interface, Std_Package.Natural_Subtype_Definition);
            Set_Mode (Interface, Iir_Out_Mode);
            Append (Last_Interface, Proc, Interface);
            Set_Implicit_Definition (Proc, Iir_Predefined_Read_Length);
         when others =>
            Set_Implicit_Definition (Proc, Iir_Predefined_Read);
      end case;
      -- Add it to the list.
      Insert_Incr (Last, Proc);

      -- Create the implicit procedure write declaration.
      Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
      Set_Identifier (Proc, Std_Names.Name_Write);
      Set_Location (Proc, Loc);
      Set_Parent (Proc, Get_Parent (Decl));
      Set_Type_Reference (Proc, Decl);
      Build_Init (Last_Interface);
      Interface := Create_Iir (File_Interface_Kind);
      Set_Identifier (Interface, Std_Names.Name_F);
      Set_Location (Interface, Loc);
      Set_Type (Interface, Type_Definition);
      Set_Mode (Interface, Iir_Inout_Mode);
      Append (Last_Interface, Proc, Interface);
      Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
      Set_Identifier (Interface, Std_Names.Name_Value);
      Set_Location (Interface, Loc);
      Set_Type (Interface, Type_Mark);
      Set_Mode (Interface, Iir_In_Mode);
      Append (Last_Interface, Proc, Interface);
      Set_Implicit_Definition (Proc, Iir_Predefined_Write);
      -- Add it to the list.
      Insert_Incr (Last, Proc);

      -- Create the implicit function endfile declaration.
      Func := Create_Iir (Iir_Kind_Implicit_Function_Declaration);
      Set_Identifier (Func, Std_Names.Name_Endfile);
      Set_Location (Func, Loc);
      Set_Parent (Proc, Get_Parent (Decl));
      Set_Type_Reference (Proc, Decl);
      Build_Init (Last_Interface);
      Interface := Create_Iir (File_Interface_Kind);
      Set_Identifier (Interface, Std_Names.Name_F);
      Set_Location (Interface, Loc);
      Set_Type (Interface, Type_Definition);
      Set_Mode (Interface, Iir_Inout_Mode);
      Append (Last_Interface, Func, Interface);
      Set_Return_Type (Func, Std_Package.Boolean_Type_Definition);
      Set_Implicit_Definition (Func, Iir_Predefined_Endfile);
      -- Add it to the list.
      Insert_Incr (Last, Func);
   end Create_Implicit_File_Primitives;

   function Create_Anonymous_Interface
     (Atype : Iir)
     return Iir_Constant_Interface_Declaration
   is
      Interface : Iir_Constant_Interface_Declaration;
   begin
      Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
      Location_Copy (Interface, Atype);
      Set_Identifier (Interface, Null_Identifier);
      Set_Mode (Interface, Iir_In_Mode);
      Set_Type (Interface, Atype);
      return Interface;
   end Create_Anonymous_Interface;

   procedure Create_Implicit_Operations
     (Decl : Iir; Is_Std_Standard : Boolean := False)
   is
      use Std_Names;
      Binary_Chain : Iir;
      Unary_Chain : Iir;
      Type_Definition : Iir;
      Last : Iir;

      procedure Add_Operation
        (Name : Name_Id;
         Def : Iir_Predefined_Functions;
         Interface_Chain : Iir;
         Return_Type : Iir)
      is
         Operation : Iir_Implicit_Function_Declaration;
      begin
         Operation := Create_Iir (Iir_Kind_Implicit_Function_Declaration);
         Location_Copy (Operation, Decl);
         Set_Parent (Operation, Get_Parent (Decl));
         Set_Interface_Declaration_Chain (Operation, Interface_Chain);
         Set_Type_Reference (Operation, Decl);
         Set_Return_Type (Operation, Return_Type);
         Set_Implicit_Definition (Operation, Def);
         Set_Identifier (Operation, Name);
         Insert_Incr (Last, Operation);
      end Add_Operation;

      procedure Add_Relational (Name : Name_Id; Def : Iir_Predefined_Functions)
      is
      begin
         Add_Operation
           (Name, Def, Binary_Chain, Std_Package.Boolean_Type_Definition);
      end Add_Relational;

      procedure Add_Binary (Name : Name_Id; Def : Iir_Predefined_Functions) is
      begin
         Add_Operation (Name, Def, Binary_Chain, Type_Definition);
      end Add_Binary;

      procedure Add_Unary (Name : Name_Id; Def : Iir_Predefined_Functions) is
      begin
         Add_Operation (Name, Def, Unary_Chain, Type_Definition);
      end Add_Unary;

      procedure Add_Shift_Operators
      is
         Inter_Chain : Iir_Constant_Interface_Declaration;
         Inter_Int : Iir;
      begin
         Inter_Chain := Create_Anonymous_Interface (Type_Definition);

         Inter_Int := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
         Location_Copy (Inter_Int, Decl);
         Set_Identifier (Inter_Int, Null_Identifier);
         Set_Mode (Inter_Int, Iir_In_Mode);
         Set_Type (Inter_Int, Std_Package.Integer_Subtype_Definition);

         Set_Chain (Inter_Chain, Inter_Int);

         Add_Operation
           (Name_Sll, Iir_Predefined_Array_Sll, Inter_Chain, Type_Definition);
         Add_Operation
           (Name_Srl, Iir_Predefined_Array_Srl, Inter_Chain, Type_Definition);
         Add_Operation
           (Name_Sla, Iir_Predefined_Array_Sla, Inter_Chain, Type_Definition);
         Add_Operation
           (Name_Sra, Iir_Predefined_Array_Sra, Inter_Chain, Type_Definition);
         Add_Operation
           (Name_Rol, Iir_Predefined_Array_Rol, Inter_Chain, Type_Definition);
         Add_Operation
           (Name_Ror, Iir_Predefined_Array_Ror, Inter_Chain, Type_Definition);
      end Add_Shift_Operators;
   begin
      Last := Decl;

      Type_Definition := Get_Base_Type (Get_Type (Decl));
      if Get_Kind (Type_Definition) /= Iir_Kind_File_Type_Definition then
         Unary_Chain := Create_Anonymous_Interface (Type_Definition);
         Binary_Chain := Create_Anonymous_Interface (Type_Definition);
         Set_Chain (Binary_Chain, Unary_Chain);
      end if;

      case Get_Kind (Type_Definition) is
         when Iir_Kind_Enumeration_Type_Definition =>
            Add_Relational (Name_Op_Equality, Iir_Predefined_Enum_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Enum_Inequality);
            Add_Relational (Name_Op_Greater, Iir_Predefined_Enum_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Enum_Greater_Equal);
            Add_Relational (Name_Op_Less, Iir_Predefined_Enum_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Enum_Less_Equal);

         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Array_Subtype_Definition =>
            declare
               Inter_Chain : Iir;
               Element_Type : Iir;
            begin
               Add_Relational
                 (Name_Op_Equality, Iir_Predefined_Array_Equality);
               Add_Relational
                 (Name_Op_Inequality, Iir_Predefined_Array_Inequality);
               if Is_Discrete_Array (Type_Definition) then
                  Add_Relational
                    (Name_Op_Greater, Iir_Predefined_Array_Greater);
                  Add_Relational
                    (Name_Op_Greater_Equal,
                     Iir_Predefined_Array_Greater_Equal);
                  Add_Relational
                    (Name_Op_Less, Iir_Predefined_Array_Less);
                  Add_Relational
                    (Name_Op_Less_Equal, Iir_Predefined_Array_Less_Equal);
               end if;

               Element_Type :=
                 Get_Base_Type (Get_Element_Subtype (Type_Definition));

               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Array_Array_Concat,
                              Binary_Chain,
                              Type_Definition);

               Inter_Chain := Create_Anonymous_Interface (Element_Type);
               Set_Chain (Inter_Chain, Unary_Chain);
               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Element_Array_Concat,
                              Inter_Chain,
                              Type_Definition);

               Inter_Chain := Create_Anonymous_Interface (Type_Definition);
               Set_Chain (Inter_Chain,
                          Create_Anonymous_Interface (Element_Type));
               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Array_Element_Concat,
                              Inter_Chain,
                              Type_Definition);

               Inter_Chain := Create_Anonymous_Interface (Element_Type);
               Set_Chain (Inter_Chain,
                          Create_Anonymous_Interface (Element_Type));
               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Element_Element_Concat,
                              Inter_Chain,
                              Type_Definition);

               if Is_Discrete_Array (Type_Definition) then
                  if Element_Type = Std_Package.Boolean_Type_Definition then
                     Add_Unary (Name_Not, Iir_Predefined_Boolean_Array_Not);

                     Add_Binary (Name_And, Iir_Predefined_Boolean_Array_And);
                     Add_Binary (Name_Or, Iir_Predefined_Boolean_Array_Or);
                     Add_Binary (Name_Nand, Iir_Predefined_Boolean_Array_Nand);
                     Add_Binary (Name_Nor, Iir_Predefined_Boolean_Array_Nor);
                     Add_Binary (Name_Xor, Iir_Predefined_Boolean_Array_Xor);
                     if Flags.Vhdl_Std > Vhdl_87 then
                        Add_Binary
                          (Name_Xnor, Iir_Predefined_Boolean_Array_Xnor);

                        Add_Shift_Operators;
                     end if;
                  elsif Element_Type = Std_Package.Bit_Type_Definition then
                     Add_Unary (Name_Not, Iir_Predefined_Bit_Array_Not);

                     Add_Binary (Name_And, Iir_Predefined_Bit_Array_And);
                     Add_Binary (Name_Or, Iir_Predefined_Bit_Array_Or);
                     Add_Binary (Name_Nand, Iir_Predefined_Bit_Array_Nand);
                     Add_Binary (Name_Nor, Iir_Predefined_Bit_Array_Nor);
                     Add_Binary (Name_Xor, Iir_Predefined_Bit_Array_Xor);
                     if Flags.Vhdl_Std > Vhdl_87 then
                        Add_Binary (Name_Xnor, Iir_Predefined_Bit_Array_Xnor);

                        Add_Shift_Operators;
                     end if;
                  end if;
               end if;
            end;

         when Iir_Kind_Access_Type_Definition =>
            Add_Relational (Name_Op_Equality, Iir_Predefined_Access_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Access_Inequality);
            declare
               Deallocate_Proc: Iir_Implicit_Procedure_Declaration;
               Var_Interface: Iir_Variable_Interface_Declaration;
            begin
               Deallocate_Proc :=
                 Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
               Set_Identifier (Deallocate_Proc, Std_Names.Name_Deallocate);
               Set_Implicit_Definition
                 (Deallocate_Proc, Iir_Predefined_Deallocate);
               Var_Interface :=
                 Create_Iir (Iir_Kind_Variable_Interface_Declaration);
               Set_Identifier (Var_Interface, Std_Names.Name_P);
               Set_Type (Var_Interface, Type_Definition);
               Set_Mode (Var_Interface, Iir_Inout_Mode);
               --Set_Purity_State (Deallocate_Proc, Impure);
               Set_Wait_State (Deallocate_Proc, False);
               Set_Type_Reference (Deallocate_Proc, Decl);

               Set_Interface_Declaration_Chain
                 (Deallocate_Proc, Var_Interface);
               Insert_Incr (Last, Deallocate_Proc);
            end;

         when Iir_Kind_Record_Type_Definition =>
            Add_Relational (Name_Op_Equality, Iir_Predefined_Record_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Record_Inequality);

         when Iir_Kind_Integer_Type_Definition =>
            Add_Relational (Name_Op_Equality, Iir_Predefined_Integer_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Integer_Inequality);
            Add_Relational (Name_Op_Greater, Iir_Predefined_Integer_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Integer_Greater_Equal);
            Add_Relational (Name_Op_Less, Iir_Predefined_Integer_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Integer_Less_Equal);

            Add_Binary (Name_Op_Plus, Iir_Predefined_Integer_Plus);
            Add_Binary (Name_Op_Minus, Iir_Predefined_Integer_Minus);

            Add_Unary (Name_Op_Minus, Iir_Predefined_Integer_Negation);
            Add_Unary (Name_Op_Plus, Iir_Predefined_Integer_Identity);

            Add_Binary (Name_Op_Mul, Iir_Predefined_Integer_Mul);
            Add_Binary (Name_Op_Div, Iir_Predefined_Integer_Div);
            Add_Binary (Name_Mod, Iir_Predefined_Integer_Mod);
            Add_Binary (Name_Rem, Iir_Predefined_Integer_Rem);

            Add_Unary (Name_Abs, Iir_Predefined_Integer_Absolute);

            declare
               Inter_Chain : Iir;
            begin
               Inter_Chain := Create_Anonymous_Interface (Type_Definition);
               Set_Chain
                 (Inter_Chain,
                  Create_Anonymous_Interface (Integer_Type_Definition));
               Add_Operation (Name_Op_Exp, Iir_Predefined_Integer_Exp,
                              Inter_Chain, Type_Definition);
            end;

         when Iir_Kind_Floating_Type_Definition =>
            Add_Relational
              (Name_Op_Equality, Iir_Predefined_Floating_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Floating_Inequality);
            Add_Relational
              (Name_Op_Greater, Iir_Predefined_Floating_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Floating_Greater_Equal);
            Add_Relational
              (Name_Op_Less, Iir_Predefined_Floating_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Floating_Less_Equal);

            Add_Binary (Name_Op_Plus, Iir_Predefined_Floating_Plus);
            Add_Binary (Name_Op_Minus, Iir_Predefined_Floating_Minus);

            Add_Unary (Name_Op_Minus, Iir_Predefined_Floating_Negation);
            Add_Unary (Name_Op_Plus, Iir_Predefined_Floating_Identity);

            Add_Binary (Name_Op_Mul, Iir_Predefined_Floating_Mul);
            Add_Binary (Name_Op_Div, Iir_Predefined_Floating_Div);

            Add_Unary (Name_Abs, Iir_Predefined_Floating_Absolute);

            declare
               Inter_Chain : Iir;
            begin
               Inter_Chain := Create_Anonymous_Interface (Type_Definition);
               Set_Chain
                 (Inter_Chain,
                  Create_Anonymous_Interface (Integer_Type_Definition));
               Add_Operation (Name_Op_Exp, Iir_Predefined_Floating_Exp,
                              Inter_Chain, Type_Definition);
            end;

         when Iir_Kind_Physical_Type_Definition =>
            Add_Relational
              (Name_Op_Equality, Iir_Predefined_Physical_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Physical_Inequality);
            Add_Relational
              (Name_Op_Greater, Iir_Predefined_Physical_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Physical_Greater_Equal);
            Add_Relational
              (Name_Op_Less, Iir_Predefined_Physical_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Physical_Less_Equal);

            Add_Binary (Name_Op_Plus, Iir_Predefined_Physical_Plus);
            Add_Binary (Name_Op_Minus, Iir_Predefined_Physical_Minus);

            Add_Unary (Name_Op_Minus, Iir_Predefined_Physical_Negation);
            Add_Unary (Name_Op_Plus, Iir_Predefined_Physical_Identity);

            declare
               Inter_Chain : Iir;
            begin
               Inter_Chain := Create_Anonymous_Interface (Type_Definition);
               Set_Chain
                 (Inter_Chain,
                  Create_Anonymous_Interface (Integer_Type_Definition));
               Add_Operation (Name_Op_Mul, Iir_Predefined_Physical_Integer_Mul,
                              Inter_Chain, Type_Definition);
               Add_Operation (Name_Op_Div, Iir_Predefined_Physical_Integer_Div,
                              Inter_Chain, Type_Definition);
            end;

            declare
               Inter_Chain : Iir;
            begin
               Inter_Chain :=
                 Create_Anonymous_Interface (Integer_Type_Definition);
               Set_Chain (Inter_Chain, Unary_Chain);
               Add_Operation (Name_Op_Mul, Iir_Predefined_Integer_Physical_Mul,
                              Inter_Chain, Type_Definition);
            end;

            declare
               Inter_Chain : Iir;
            begin
               Inter_Chain := Create_Anonymous_Interface (Type_Definition);
               Set_Chain (Inter_Chain,
                          Create_Anonymous_Interface (Real_Type_Definition));
               Add_Operation (Name_Op_Mul, Iir_Predefined_Physical_Real_Mul,
                              Inter_Chain, Type_Definition);
               Add_Operation (Name_Op_Div, Iir_Predefined_Physical_Real_Div,
                              Inter_Chain, Type_Definition);
            end;

            declare
               Inter_Chain : Iir;
            begin
               Inter_Chain :=
                 Create_Anonymous_Interface (Real_Type_Definition);
               Set_Chain (Inter_Chain, Unary_Chain);
               Add_Operation (Name_Op_Mul, Iir_Predefined_Real_Physical_Mul,
                              Inter_Chain, Type_Definition);
            end;
            Add_Operation (Name_Op_Div, Iir_Predefined_Physical_Physical_Div,
                           Binary_Chain,
                           Std_Package.Universal_Integer_Type_Definition);

            Add_Unary (Name_Abs, Iir_Predefined_Physical_Absolute);

         when Iir_Kind_File_Type_Definition =>
            Create_Implicit_File_Primitives (Decl, Type_Definition);

         when Iir_Kind_Protected_Type_Declaration =>
            null;

         when others =>
            Error_Kind ("create_predefined_operations", Type_Definition);
      end case;

      if not Is_Std_Standard then
         return;
      end if;
      if Decl = Std_Package.Boolean_Type then
         Add_Binary (Name_And, Iir_Predefined_Boolean_And);
         Add_Binary (Name_Or, Iir_Predefined_Boolean_Or);
         Add_Binary (Name_Nand, Iir_Predefined_Boolean_Nand);
         Add_Binary (Name_Nor, Iir_Predefined_Boolean_Nor);
         Add_Binary (Name_Xor, Iir_Predefined_Boolean_Xor);
         if Flags.Vhdl_Std > Vhdl_87 then
            Add_Binary (Name_Xnor, Iir_Predefined_Boolean_Xnor);
         end if;
         Add_Unary (Name_Not, Iir_Predefined_Boolean_Not);
      elsif Decl = Std_Package.Bit_Type then
         Add_Binary (Name_And, Iir_Predefined_Bit_And);
         Add_Binary (Name_Or, Iir_Predefined_Bit_Or);
         Add_Binary (Name_Nand, Iir_Predefined_Bit_Nand);
         Add_Binary (Name_Nor, Iir_Predefined_Bit_Nor);
         Add_Binary (Name_Xor, Iir_Predefined_Bit_Xor);
         if Flags.Vhdl_Std > Vhdl_87 then
            Add_Binary (Name_Xnor, Iir_Predefined_Bit_Xnor);
         end if;
         Add_Unary (Name_Not, Iir_Predefined_Bit_Not);
      elsif Decl = Std_Package.Universal_Real_Type then
         declare
            Inter_Chain : Iir;
         begin
            Inter_Chain := Create_Anonymous_Interface (Type_Definition);
            Set_Chain
              (Inter_Chain,
               Create_Anonymous_Interface (Universal_Integer_Type_Definition));
            Add_Operation (Name_Op_Mul, Iir_Predefined_Universal_R_I_Mul,
                           Inter_Chain, Type_Definition);
            Add_Operation (Name_Op_Div, Iir_Predefined_Universal_R_I_Div,
                           Inter_Chain, Type_Definition);
         end;

         declare
            Inter_Chain : Iir;
         begin
            Inter_Chain :=
              Create_Anonymous_Interface (Universal_Integer_Type_Definition);
            Set_Chain (Inter_Chain, Unary_Chain);
            Add_Operation (Name_Op_Mul, Iir_Predefined_Universal_I_R_Mul,
                           Inter_Chain, Type_Definition);
         end;
      end if;
   end Create_Implicit_Operations;

   procedure Sem_Type_Declaration (Decl: Iir)
   is
      Def: Iir;
      Inter : Name_Interpretation_Type;
      Old_Decl : Iir;
      St_Decl : Iir_Subtype_Declaration;
      Bt_Def : Iir;
   begin
      --  Check if DECL complete a previous incomplete type declaration.
      Inter := Get_Interpretation (Get_Identifier (Decl));
      if Valid_Interpretation (Inter)
        and then Is_In_Current_Declarative_Region (Inter)
      then
         Old_Decl := Get_Declaration (Inter);
         if Get_Kind (Old_Decl) /= Iir_Kind_Type_Declaration
           or else Get_Kind (Get_Type (Old_Decl)) /=
           Iir_Kind_Incomplete_Type_Definition
         then
            Old_Decl := Null_Iir;
         end if;
      else
         Old_Decl := Null_Iir;
      end if;

      if Old_Decl = Null_Iir then
         if Get_Kind (Decl) = Iir_Kind_Type_Declaration then
            --  This is necessary at least for enumeration type definition.
            Sem_Scopes.Add_Name (Decl);
         end if;
      else
         --  This is a way to prevent:
         --    type a;
         --    type a is access a;
         --  which is non-sense.
         Set_Visible_Flag (Old_Decl, False);
      end if;

      -- Check the definition of the type.
      Def := Get_Type (Decl);
      if Def = Null_Iir then
         --  Incomplete type declaration
         Def := Create_Iir (Iir_Kind_Incomplete_Type_Definition);
         Location_Copy (Def, Decl);
         Set_Type (Decl, Def);
         Set_Signal_Type_Flag (Def, True);
         Set_Type_Declarator (Def, Decl);
         Set_Visible_Flag (Decl, True);
         Set_Incomplete_Type_List (Def, Create_Iir_List);
      else
         Def := Sem_Type_Definition (Def, Decl);
         if Def /= Null_Iir then
            case Get_Kind (Def) is
               when Iir_Kind_Integer_Subtype_Definition
                 | Iir_Kind_Floating_Subtype_Definition
                 | Iir_Kind_Physical_Subtype_Definition
                 | Iir_Kind_Array_Subtype_Definition =>
                  Bt_Def := Get_Base_Type (Def);
                  St_Decl := Create_Iir (Iir_Kind_Subtype_Declaration);
                  Location_Copy (St_Decl, Decl);
                  Set_Identifier (St_Decl, Get_Identifier (Decl));
                  Set_Type (St_Decl, Def);
                  Set_Type_Declarator (Def, St_Decl);
                  Set_Chain (St_Decl, Get_Chain (Decl));
                  Set_Chain (Decl, St_Decl);

                  Set_Type (Decl, Bt_Def);
                  Set_Type_Declarator (Bt_Def, Decl);
                  Set_Subtype_Definition (Decl, Def);

                  if Old_Decl = Null_Iir then
                     Sem_Scopes.Add_Name (St_Decl);
                  else
                     Replace_Name (Get_Identifier (Decl), Old_Decl, St_Decl);
                     Set_Type_Declarator (Get_Type (Old_Decl), St_Decl);
                  end if;

                  Sem_Scopes.Name_Visible (St_Decl);

                  Sem_Scopes.Add_Visible_Type (Decl);

                  --  The implicit subprogram will be added in the
                  -- scope just after.
                  Create_Implicit_Operations (Decl, False);

               when Iir_Kind_Enumeration_Type_Definition
                 | Iir_Kind_Array_Type_Definition
                 | Iir_Kind_Record_Type_Definition
                 | Iir_Kind_Access_Type_Definition
                 | Iir_Kind_File_Type_Definition =>
                  St_Decl := Null_Iir;
                  Set_Type_Declarator (Def, Decl);

                  Sem_Scopes.Name_Visible (Decl);
                  Sem_Scopes.Add_Visible_Type (Decl);

                  --  The implicit subprogram will be added in the
                  -- scope just after.
                  Create_Implicit_Operations (Decl, False);

               when Iir_Kind_Protected_Type_Declaration =>
                  Set_Type_Declarator (Def, Decl);
                  Sem_Scopes.Add_Visible_Type (Decl);
                  St_Decl := Null_Iir;
                  --  No implicit subprograms.

               when others =>
                  Error_Kind ("sem_type_declaration", Def);
            end case;

            if Old_Decl /= Null_Iir then
               --  Complete the type definition.
               declare
                  List : Iir_List;
                  El : Iir;
                  Old_Def : Iir;
               begin
                  Old_Def := Get_Type (Old_Decl);
                  Set_Signal_Type_Flag (Old_Def, Get_Signal_Type_Flag (Def));
                  List := Get_Incomplete_Type_List (Old_Def);
                  for I in Natural loop
                     El := Get_Nth_Element (List, I);
                     exit when El = Null_Iir;
                     Set_Designated_Type (El, Def);
                  end loop;
                  --  Complete the incomplete_type_definition node
                  --  (set type_declarator and base_type).

                  Set_Base_Type (Old_Def, Get_Base_Type (Def));
                  if St_Decl = Null_Iir then
                     Set_Type_Declarator (Old_Def, Decl);
                     Replace_Name (Get_Identifier (Decl), Old_Decl, Decl);
                  end if;
               end;
            end if;
         end if;
      end if;
   end Sem_Type_Declaration;

   procedure Sem_Subtype_Declaration (Decl: Iir)
   is
      Def: Iir;
      Res: Iir;
   begin
      --  Real hack to skip subtype declarations of anonymous type decls.
      if Get_Visible_Flag (Decl) then
         return;
      end if;

      Sem_Scopes.Add_Name (Decl);

      -- Check the definition of the type.
      Def := Sem_Subtype_Indication (Get_Type (Decl));
      if Def = Null_Iir then
         return;
      end if;

      if not Is_Anonymous_Type_Definition (Def) then
         case Get_Kind (Def) is
            when Iir_Kind_Integer_Subtype_Definition
              | Iir_Kind_Floating_Subtype_Definition
              | Iir_Kind_Enumeration_Subtype_Definition
              | Iir_Kind_Physical_Subtype_Definition =>
               -- no limits, makes an alias.
               Res := Create_Iir (Get_Kind (Def));
               Set_Range_Constraint (Res, Get_Range_Constraint (Def));
               Set_Resolution_Function (Res, Get_Resolution_Function (Def));
            when Iir_Kind_Enumeration_Type_Definition =>
               --  makes an alias.
               Res := Create_Iir (Iir_Kind_Enumeration_Subtype_Definition);
               Set_Type_Mark (Res, Def);
               Set_Range_Constraint (Res, Get_Range_Constraint (Def));
            when Iir_Kind_Access_Subtype_Definition
              | Iir_Kind_Access_Type_Definition =>
               --  Make an alias.
               Res := Create_Iir (Iir_Kind_Access_Subtype_Definition);
            when Iir_Kind_Array_Type_Definition
              | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
               Res :=
                 Create_Iir (Iir_Kind_Unconstrained_Array_Subtype_Definition);
               Set_Type_Staticness (Res, Get_Type_Staticness (Def));
               if Get_Kind (Def) /= Iir_Kind_Array_Type_Definition then
                  Set_Resolution_Function
                    (Res, Get_Resolution_Function (Def));
               end if;
               Set_Resolved_Flag (Res, Get_Resolved_Flag (Def));
               Set_Type_Mark (Res, Def);
            when Iir_Kind_Array_Subtype_Definition =>
               Res := Create_Iir (Iir_Kind_Array_Subtype_Definition);
               Set_Resolution_Function (Res, Get_Resolution_Function (Def));
               Set_Resolved_Flag (Res, Get_Resolved_Flag (Def));
               Set_Type_Mark (Res, Def);
               Set_Index_Subtype_List (Res, Get_Index_Subtype_List (Def));
               Set_Element_Subtype (Res, Get_Element_Subtype (Def));
            when Iir_Kind_Record_Type_Definition
              | Iir_Kind_Record_Subtype_Definition =>
               Res := Create_Iir (Iir_Kind_Record_Subtype_Definition);
               Set_Type_Staticness (Res, Get_Type_Staticness (Def));
               if Get_Kind (Def) /= Iir_Kind_Record_Type_Definition then
                  Set_Resolution_Function
                    (Res, Get_Resolution_Function (Def));
               end if;
               Set_Resolved_Flag (Res, Get_Resolved_Flag (Def));
            when others =>
               --  FIXME: todo
               Error_Kind ("sem_subtype_declaration", Def);
         end case;
         Location_Copy (Res, Decl);
         Set_Base_Type (Res, Get_Base_Type (Def));
         Set_Type_Staticness (Res, Get_Type_Staticness (Def));
         Set_Signal_Type_Flag (Res, Get_Signal_Type_Flag (Def));
         Def := Res;
      end if;
      Set_Type (Decl, Def);
      Set_Type_Declarator (Def, Decl);
      Name_Visible (Decl);
   end Sem_Subtype_Declaration;

   --  If DECL is a constant declaration, and there is already a constant
   --  declaration in the current scope with the same name, then return it.
   --  Otherwise, return NULL.
   function Get_Deferred_Constant (Decl : Iir) return Iir
   is
      Deferred_Const : Iir;
      Interp : Name_Interpretation_Type;
   begin
      if Get_Kind (Decl) /= Iir_Kind_Constant_Declaration then
         return Null_Iir;
      end if;
      Interp := Get_Interpretation (Get_Identifier (Decl));
      if not Valid_Interpretation (Interp) then
         return Null_Iir;
      end if;
      if not Is_In_Current_Declarative_Region (Interp) then
         return Null_Iir;
      end if;
      Deferred_Const := Get_Declaration (Interp);
      if Get_Kind (Deferred_Const) /= Iir_Kind_Constant_Declaration then
         return Null_Iir;
      end if;
      --  LRM93 4.3.1.1
      --  The corresponding full constant declaration, which defines the value
      --  of the constant, must appear in the body of the package.
      if Get_Kind (Get_Library_Unit (Get_Current_Design_Unit))
        /= Iir_Kind_Package_Body
      then
         Error_Msg_Sem
           ("full constant declaration must appear in package body", Decl);
      end if;
      return Deferred_Const;
   end Get_Deferred_Constant;

   procedure Sem_Object_Declaration (Decl: Iir; Parent : Iir)
   is
      Atype: Iir;
      Default_Value : Iir;
      Proxy : Iir;
      Deferred_Const : Iir;
      Staticness : Iir_Staticness;
   begin
      Deferred_Const := Get_Deferred_Constant (Decl);

      --  Semantize type and default value:
      Atype := Get_Type (Decl);
      if Get_Kind (Atype) /= Iir_Kind_Proxy then
         Atype := Sem_Subtype_Indication (Atype);
         if Atype = Null_Iir then
            Set_Type (Decl, Null_Iir);
            return;
         end if;
      end if;

      if Deferred_Const = Null_Iir then
         Sem_Scopes.Add_Name (Decl);
      end if;

      if Get_Kind (Atype) = Iir_Kind_Proxy then
         Proxy := Get_Proxy (Atype);
         Default_Value := Get_Default_Value (Proxy);
         Atype := Get_Type (Proxy);
         if Atype = Null_Iir then
            return;
         end if;
         Proxy := Get_Type (Decl);
         Free_Iir (Proxy);
      else
         Default_Value := Get_Default_Value (Decl);
         if Default_Value /= Null_Iir then
            declare
               Val : Iir;
            begin
               Val := Sem_Expression (Default_Value, Atype);
               if Val /= Null_Iir then
                  Default_Value := Val;
               end if;
            end;
         end if;
      end if;
      Set_Type (Decl, Atype);
      Set_Default_Value (Decl, Default_Value);
      Set_Base_Name (Decl, Decl);
      Set_Name_Staticness (Decl, Locally);
      Set_Visible_Flag (Decl, True);

      --  LRM93 2.6
      --  The subtype indication given in the full declaration of the deferred
      --  constant must conform to that given in the deferred constant
      --  declaration.
      if Deferred_Const /= Null_Iir
        and then not Are_Trees_Equal (Get_Type (Decl),
                                      Get_Type (Deferred_Const))
      then
         Error_Msg_Sem
           ("subtype indication doesn't conform with the deferred constant",
            Decl);
      end if;

      --  LRM 4.3.1.3
      --  It is an error if a variable declaration declares a variable that is
      --  of a file type.
      --
      --  LRM 4.3.1.1
      --  It is an error if a constant declaration declares a constant that is
      --  of a file type, or an access type, or a composite type which has
      --  subelement that is a file type of an access type.
      --
      --  LRM 4.3.1.2
      --  It is an error if a signal declaration declares a signal that is of
      --  a file type [or an access type].
      case Get_Kind (Atype) is
         when Iir_Kind_File_Type_Definition =>
            Error_Msg_Sem (Disp_Node (Decl) & " cannot be of type file", Decl);
         when others =>
            if Get_Kind (Decl) /= Iir_Kind_Variable_Declaration
              and then Get_Signal_Type_Flag (Atype) = False
            then
               Error_Msg_Sem ("type of " & Disp_Node (Decl)
                              & " cannot be " & Disp_Node (Atype), Decl);
               Error_Msg_Sem ("(" & Disp_Node (Atype)
                              & " has an access subelement)", Decl);
            end if;
      end case;

      case Get_Kind (Decl) is
         when Iir_Kind_Constant_Declaration =>
            --  LRM93 4.3.1.1
            --  If the assignment symbol ":=" followed by an expression is not
            --  present in a constant declaration, then the declaration
            --  declares a deferred constant.
            --  Such a constant declaration may only appear in a package
            --  declaration.
            if Deferred_Const /= Null_Iir then
               Set_Deferred_Declaration (Decl, Deferred_Const);
               Set_Deferred_Declaration (Deferred_Const, Decl);
            end if;
            if Default_Value = Null_Iir then
               if Deferred_Const /= Null_Iir then
                  Error_Msg_Sem
                    ("full constant declaration must have a default value",
                     Decl);
               else
                  Set_Deferred_Declaration_Flag (Decl, True);
               end if;
               if Get_Kind (Parent) /= Iir_Kind_Package_Declaration then
                  Error_Msg_Sem ("a constant must have a default value", Decl);
               end if;
               Set_Expr_Staticness (Decl, Globally);
            else
               --  LRM93 7.4.1: a locally static primary is defined:
               --  A constant (other than deferred constant) explicitly
               --  declared by a constant declaration and initialized
               --  with a locally static expression.
               --  Note: the staticness of the full declaration may be locally.
               if False and Deferred_Const /= Null_Iir then
                  --  This is a deferred constant.
                  Staticness := Globally;
               else
                  Staticness := Min (Get_Expr_Staticness (Default_Value),
                                     Get_Type_Staticness (Atype));
                  --  LRM 7.4.2 (Globally static primaries)
                  --  5. a constant
                  if Staticness < Globally then
                     Staticness := Globally;
                  end if;
               end if;
               Set_Expr_Staticness (Decl, Staticness);

               if Staticness = Locally then
                  Set_Default_Value
                    (Decl, Eval_Expr_Check (Default_Value, Atype));
               end if;
            end if;
         when Iir_Kind_Signal_Declaration =>
            --  LRM93 4.3.1.2
            --  It is also an error if a guarded signal of a
            --  scalar type is neither a resolved signal nor a
            --  subelement of a resolved signal.
            if Get_Signal_Kind (Decl) /= Iir_No_Signal_Kind
              and then not Get_Resolved_Flag (Atype)
            then
               Error_Msg_Sem
                 ("guarded " & Disp_Node (Decl) & " must be resolved", Decl);
            end if;
            Set_Expr_Staticness (Decl, None);
            Set_Has_Disconnect_Flag (Decl, False);
         when Iir_Kind_Variable_Declaration =>
            Set_Expr_Staticness (Decl, None);
         when others =>
            Error_Kind ("sem_object_declaration", Decl);
      end case;

      case Get_Kind (Decl) is
         when Iir_Kind_Constant_Declaration =>
            --  LRM93 3.2.1.1
            --  For a constant declared by an object declaration, the index
            --  ranges are defined by the initial value, if the subtype of the
            --  constant is unconstrained; otherwise they are defined by this
            --  subtype.
            --if Default_Value = Null_Iir
            --  and then not Sem_Is_Constrained (Atype)
            --then
            --   Error_Msg_Sem ("constant declaration of unconstrained "
            --                  & Disp_Node (Atype) & " is not allowed", Decl);
            --end if;
            null;
            --if Deferred_Const = Null_Iir then
            --   Name_Visible (Decl);
            --end if;

         when Iir_Kind_Variable_Declaration
           | Iir_Kind_Signal_Declaration =>
            --  LRM93 3.2.1.1
            --  For a variable or signal declared by an object declaration, the
            --  subtype indication of the corressponding object declaration
            --  must define a constrained array subtype.
            if not Sem_Is_Constrained (Atype) then
               Error_Msg_Sem
                 ("declaration of " & Disp_Node (Decl)
                  & " with unconstrained " & Disp_Node (Atype)
                  & " is not allowed", Decl);
               if Default_Value /= Null_Iir then
                  Error_Msg_Sem ("(even with a default value)", Decl);
               end if;
            end if;

         when others =>
            Error_Kind ("sem_object_declaration(2)", Decl);
      end case;
   end Sem_Object_Declaration;

   procedure Sem_File_Declaration (Decl: Iir_File_Declaration)
   is
      Atype: Iir;
      Logical_Name: Iir;
      Open_Kind : Iir;
   begin
      Sem_Scopes.Add_Name (Decl);
      Set_Expr_Staticness (Decl, None);
      Set_Base_Name (Decl, Decl);

      -- Try to find a type.
      Atype := Get_Type (Decl);
      if Get_Kind (Atype) = Iir_Kind_Proxy then
         Atype := Get_Type (Get_Proxy (Atype));
         Free_Iir (Get_Type (Decl));
      else
         Atype := Sem_Subtype_Indication (Get_Type (Decl));
         if Atype = Null_Iir then
            return;
         end if;
      end if;
      Set_Type (Decl, Atype);

      --  LRM93 4.3.1.4
      --  The subtype indication of a file declaration must define a file
      --  subtype.
      if Get_Kind (Atype) /= Iir_Kind_File_Type_Definition then
         Error_Msg_Sem ("file subtype expected for a file declaration", Decl);
         return;
      end if;

      Logical_Name := Get_File_Logical_Name (Decl);
      --  LRM93 4.3.1.4
      --  The file logical name must be an expression of predefined type
      --  STRING.
      if Logical_Name /= Null_Iir then
         Logical_Name := Sem_Expression (Logical_Name, String_Type_Definition);
         if Logical_Name /= Null_Iir then
            Check_Read (Logical_Name);
            Set_File_Logical_Name (Decl, Logical_Name);
         end if;
      end if;

      Open_Kind := Get_File_Open_Kind (Decl);
      if Open_Kind /= Null_Iir then
         Open_Kind := Sem_Expression
           (Open_Kind, File_Open_Kind_Type_Definition);
         if Open_Kind /= Null_Iir then
            Check_Read (Open_Kind);
            Set_File_Open_Kind (Decl, Open_Kind);
         end if;
      else
         --  LRM93 4.3.1.4
         --  If a file open kind expression is not included in the file open
         --  information of a given file declaration, then the default value
         --  of READ_MODE is used during elaboration of the file declaration.
         --
         --  LRM87 4.3.1.4
         --  The default mode is IN, if no mode is specified.
         if Get_Mode (Decl) = Iir_Unknown_Mode then
            if Flags.Vhdl_Std = Vhdl_87 then
               Set_Mode (Decl, Iir_In_Mode);
            else
               Set_File_Open_Kind (Decl, File_Open_Kind_Read_Mode);
            end if;
         end if;
      end if;
      Name_Visible (Decl);

      --  LRM 93 2.2
      --  If a pure function is the parent of a given procedure, then
      --  that procedure must not contain a reference to an explicitly
      --  declared file object [...]
      --
      --  A pure function must not contain a reference to an explicitly
      --  declared file.

      --  Note: this check is also performed when a file is referenced.
      --    But a file can be declared without being explicitly referenced.
      if Flags.Vhdl_Std > Vhdl_93c then
         declare
            Parent : Iir;
            Spec : Iir;
         begin
            Parent := Get_Parent (Decl);
            case Get_Kind (Parent) is
               when Iir_Kind_Function_Body =>
                  Spec := Get_Subprogram_Specification (Parent);
                  if Get_Pure_Flag (Spec) then
                     Error_Msg_Sem
                       ("cannot declare a file in a pure function", Decl);
                  end if;
               when Iir_Kind_Procedure_Body =>
                  Spec := Get_Subprogram_Specification (Parent);
                  Set_Purity_State (Spec, Impure);
                  Set_Impure_Depth (Parent, Iir_Depth_Impure);
               when Iir_Kind_Function_Declaration
                 | Iir_Kind_Procedure_Declaration =>
                  Error_Kind ("sem_file_declaration", Parent);
               when others =>
                  null;
            end case;
         end;
      end if;
   end Sem_File_Declaration;

   procedure Sem_Attribute_Declaration (Decl: Iir_Attribute_Declaration)
   is
      A_Type : Iir;
      Ident : Name_Id;
   begin
      --  LRM93 4.4
      --  The identifier is said to be the designator of the attribute.
      Ident := Get_Identifier (Decl);
      if Ident in Std_Names.Name_Id_Attributes then
         Warning_Msg_Sem ("predefined attribute """ & Name_Table.Image (Ident)
                          & """ overriden", Decl);
      end if;
      Sem_Scopes.Add_Name (Decl);

      A_Type := Sem_Subtype_Indication (Get_Type (Decl));
      if A_Type = Null_Iir then
         return;
      end if;
      --  LRM93 4.4
      --  The type mark must denote a subtype that is neither an access type
      --  nor a file type.
      --  The subtype need not be constrained.
      case Get_Kind (A_Type) is
         when Iir_Kind_Access_Type_Definition
           | Iir_Kind_Access_Subtype_Definition
           | Iir_Kind_File_Type_Definition =>
            --  FIXME: | Iir_Kind_File_Subtype_Definition =>
            Error_Msg_Sem
              ("attribute type can't be an access type nor a file type", Decl);
         when others =>
            null;
      end case;
      Set_Type (Decl, A_Type);
      Name_Visible (Decl);
   end Sem_Attribute_Declaration;

   procedure Sem_Group_Template_Declaration
     (Decl : Iir_Group_Template_Declaration)
   is
   begin
      Sem_Scopes.Add_Name (Decl);
      Sem_Scopes.Name_Visible (Decl);
   end Sem_Group_Template_Declaration;

   procedure Sem_Component_Declaration (Component: Iir_Component_Declaration)
   is
   begin
      Sem_Scopes.Add_Name (Component);

      --  LRM 10.1 Declarative region
      --  6. A component declaration.
      Open_Declarative_Region;

      Sem_Interface_Chain (Get_Generic_Chain (Component), Interface_Generic);
      Sem_Interface_Chain (Get_Port_Chain (Component), Interface_Port);

      Close_Declarative_Region;

      Name_Visible (Component);
   end Sem_Component_Declaration;

   procedure Sem_Object_Alias_Declaration (Alias: Iir_Object_Alias_Declaration)
   is
      N_Type: Iir;
      N_Name: Iir;
      Name_Type : Iir;
   begin
      Sem_Scopes.Add_Name (Alias);

      --  FIXME: currently only object alias are handled.
      Sem_Name (Get_Name (Alias));
      N_Name := Get_Named_Entity (Get_Name (Alias));
      if N_Name = Error_Mark then
         return;
      end if;
      --  FIXME: overload list ?

      Name_Visible (Alias);

      Set_Name (Alias, N_Name);
      case Get_Kind (N_Name) is
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Slice_Name
           | Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Name
           | Iir_Kind_Selected_Element =>
            Set_Base_Name (Alias, Get_Base_Name (N_Name));
         when others =>
            Error_Msg_Sem ("can only alias named object", Alias);
            return;
      end case;

      --  LRM93 4.3.3.1 Object Aliases.
      --  1. A signature may not appear in a declaration of an object alias.
      -- FIXME: todo.
      --
      --  2. The name must be a static name that denotes an object.
      if Get_Name_Staticness (N_Name) < Globally then
         Error_Msg_Sem ("aliased name must be a static name", Alias);
      end if;

      --  LRM93 4.3.3.1
      --  The base type of the name specified in an alias declaration must be
      --  the same as the base type of the type mark in the subtype indication
      --  (if the subtype indication is present);
      Name_Type := Get_Type (N_Name);
      N_Type := Get_Type (Alias);
      if N_Type = Null_Iir then
         Set_Type (Alias, Name_Type);
         N_Type := Name_Type;
      else
         N_Type := Sem_Subtype_Indication (N_Type);
         if N_Type /= Null_Iir then
            Set_Type (Alias, N_Type);
            if Get_Base_Type (N_Type) /= Get_Base_Type (Name_Type) then
               Error_Msg_Sem ("base type of aliased name and name mismatch",
                              Alias);
            end if;
         end if;
      end if;

      --  LRM93 4.3.3.1
      --  This type must not be a multi-dimensional array type.
      if Get_Kind (N_Type) in Iir_Kinds_Array_Type_Definition then
         if not Is_Unidim_Array_Type (N_Type) then
            Error_Msg_Sem
              ("aliased name must not be a multi-dimensional array type",
               Alias);
         end if;
         if Get_Type_Staticness (N_Type) = Locally
           and then Get_Type_Staticness (Name_Type) = Locally
           and then Eval_Discrete_Type_Length
           (Get_Nth_Element (Get_Index_Subtype_List (N_Type), 0))
           /= Eval_Discrete_Type_Length
           (Get_Nth_Element (Get_Index_Subtype_List (Name_Type), 0))
         then
            Error_Msg_Sem ("number of elements not matching in type and name",
                           Alias);
         end if;
      end if;

      Set_Name_Staticness (Alias, Get_Name_Staticness (N_Name));
      Set_Expr_Staticness (Alias, Get_Expr_Staticness (N_Name));
   end Sem_Object_Alias_Declaration;

   function Signature_Match (N_Entity : Iir; Sig : Iir_Signature)
                            return Boolean
   is
      List : Iir_List;
      Inter : Iir;
      El : Iir;
   begin
      List := Get_Type_Marks_List (Sig);
      case Get_Kind (N_Entity) is
         when Iir_Kind_Enumeration_Literal =>
            --  LRM93 2.3.2  Signatures
            --  * Similarly, a signature is said to match the parameter and
            --    result type profile of a given enumeration literal if
            --    the signature matches the parameter and result type profile
            --    of the subprogram equivalent to the enumeration literal,
            --    defined in Section 3.1.1
            return List = Null_Iir_List
              and then Get_Type (N_Entity) = Get_Return_Type (Sig);
         when Iir_Kind_Function_Declaration
           | Iir_Kind_Implicit_Function_Declaration =>
            --  LRM93 2.3.2  Signatures
            --  * if the reserved word RETURN is present, the subprogram is
            --    a function and the base type of the type mark following
            --    the reserved word in the signature is the same as the base
            --    type of the return type of the function, [...]
            if Get_Return_Type (Sig) /=
              Get_Base_Type (Get_Return_Type (N_Entity))
            then
               return False;
            end if;
         when Iir_Kind_Procedure_Declaration
           | Iir_Kind_Implicit_Procedure_Declaration =>
            --  LRM93 2.3.2  Signatures
            --  * [...] or the reserved word RETURN is absent and the
            --    subprogram is a procedure.
            if Get_Return_Type (Sig) /= Null_Iir then
               return False;
            end if;
         when others =>
            --  LRM93 2.3.2  Signatures
            --  A signature distinguishes between overloaded subprograms and
            --  overloaded enumeration literals based on their parameter
            --  and result type profiles.
            return False;
      end case;

      --  LRM93 2.3.2  Signature
      --  * the number of type marks prior the reserved word RETURN, if any,
      --    matches the number of formal parameters of the subprogram;
      --  * at each parameter position, the base type denoted by the type
      --    mark of the signature is the same as the base type of the
      --    corresponding formal parameter of the subprogram; [and finally, ]
      Inter := Get_Interface_Declaration_Chain (N_Entity);
      if List = Null_Iir_List then
         return Inter = Null_Iir;
      end if;
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         if El = Null_Iir and Inter = Null_Iir then
            return True;
         end if;
         if El = Null_Iir or Inter = Null_Iir then
            return False;
         end if;
         if Get_Base_Type (Get_Type (Inter)) /= El then
            return False;
         end if;
         Inter := Get_Chain (Inter);
      end loop;
      --  Avoid a spurious warning.
      return False;
   end Signature_Match;

   --  Extract from NAME the named entity whose profile matches with SIG.
   function Sem_Signature (Name : Iir; Sig : Iir_Signature) return Iir
   is
      Res : Iir;
      El : Iir;
      List : Iir_List;
      Error : Boolean;
   begin
      --  Sem signature.
      List := Get_Type_Marks_List (Sig);
      if List /= Null_Iir_List then
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            El := Find_Declaration (El, Decl_Type);
            if El /= Null_Iir then
               Replace_Nth_Element (List, I, Get_Base_Type (El));
            end if;
         end loop;
      end if;
      El := Get_Return_Type (Sig);
      if El /= Null_Iir then
         El := Find_Declaration (El, Decl_Type);
         if El /= Null_Iir then
            Set_Return_Type (Sig, Get_Base_Type (El));
         end if;
      end if;

      Res := Null_Iir;
      Error := False;
      if Is_Overload_List (Name) then
         for I in Natural loop
            El := Get_Nth_Element (Get_Overload_List (Name), I);
            exit when El = Null_Iir;
            if Signature_Match (El, Sig) then
               if Res = Null_Iir then
                  Res := El;
               else
                  Error := True;
                  Error_Msg_Sem
                    ("cannot resolve signature, many matching subprograms:",
                     Sig);
                  Error_Msg_Sem ("found: " & Disp_Node (Res), Res);
               end if;
               if Error then
                  Error_Msg_Sem ("found: " & Disp_Node (El), El);
               end if;
            end if;
         end loop;
      else
         if Signature_Match (Name, Sig) then
            Res := Name;
         end if;
      end if;

      if Error then
         return Null_Iir;
      end if;
      if Res = Null_Iir then
         Error_Msg_Sem
           ("cannot resolve signature, no matching subprogram", Sig);
      end if;
      return Res;
   end Sem_Signature;

   procedure Sem_Non_Object_Alias_Declaration
     (Alias : Iir_Non_Object_Alias_Declaration)
   is
      Name : Iir;
      Sig : Iir_Signature;
      N_Entity : Iir;
   begin
      Name := Get_Name (Alias);
      Sem_Name (Name);
      N_Entity := Get_Named_Entity (Name);
      if N_Entity = Error_Mark then
         return;
      end if;

      Sig := Get_Signature (Alias);
      if Is_Overload_List (N_Entity) then
         if Sig = Null_Iir then
            Error_Msg_Sem
              ("signature required for alias of a subprogram", Alias);
            return;
         end if;
      end if;

      if Sig /= Null_Iir then
         N_Entity := Sem_Signature (N_Entity, Sig);
         Set_Name (Alias, N_Entity);
         if N_Entity /= Null_Iir then
            Add_Name (N_Entity, Get_Identifier (Alias), False);
            Set_Visible_Flag (Alias, True);
         end if;
      else
         case Get_Kind (N_Entity) is
            when Iir_Kind_Function_Declaration
              | Iir_Kind_Implicit_Function_Declaration
              | Iir_Kind_Procedure_Declaration
              | Iir_Kind_Implicit_Procedure_Declaration =>
               --  LRM93 4.3.3.2  Non-Object Aliases
               --  2.  A signature is required if the name denotes a subprogram
               --      (including an operator) or enumeration literal.
               Error_Msg_Sem ("signature required for subprogram", Alias);
               return;
            when Iir_Kind_Enumeration_Literal =>
               Error_Msg_Sem ("signature required for enumeration literal",
                              Alias);
               return;
            when Iir_Kind_Type_Declaration =>
               declare
                  Def : Iir;
                  Last : Iir;
                  El : Iir;
                  Enum_List : Iir_Enumeration_Literal_List;

                  procedure Add_Implicit_Alias (Decl : Iir)
                  is
                     N_Alias : Iir_Non_Object_Alias_Declaration;
                  begin
                     N_Alias :=
                       Create_Iir (Iir_Kind_Non_Object_Alias_Declaration);
                     Location_Copy (N_Alias, Alias);
                     Set_Identifier (N_Alias, Get_Identifier (Decl));
                     Set_Name (N_Alias, Decl);

                     Add_Name (El, Get_Identifier (El), False);
                     Set_Visible_Flag (N_Alias, True);

                     --  Append in the declaration chain.
                     Set_Chain (N_Alias, Get_Chain (Last));
                     Set_Chain (Last, N_Alias);
                     Last := N_Alias;
                  end Add_Implicit_Alias;
               begin
                  Def := Get_Type (N_Entity);
                  Last := Alias;
                  if Get_Kind (Def) = Iir_Kind_Enumeration_Type_Definition
                  then
                     --  LRM93 4.3.3.2  Non-Object Aliases
                     --  3.  If the name denotes an enumeration type, then one
                     --      implicit alias declaration for each of the
                     --      literals of the type immediatly follows the alias
                     --      declaration for the enumeration type; [...]
                     Enum_List := Get_Enumeration_Literal_List (Def);
                     for I in Natural loop
                        El := Get_Nth_Element (Enum_List, I);
                        exit when El = Null_Iir;
                        --  LRM93 4.3.3.2  Non-Object Aliases
                        --      [...] each such implicit declaration has, as
                        --      its alias designator, the simple name or
                        --      character literal of the literal, and has,
                        --      as its name, a name constructed
                        --      by taking the name of the alias for the
                        --      enumeration type and substituting the simple
                        --      name or character literal being aliased for
                        --      the simple name of the type.
                        --      Each implicit alias has a signature that
                        --      matches the parameter and result type profile
                        --      of the literal being aliased.
                        Add_Implicit_Alias (El);
                     end loop;
                  end if;

                  --  LRM93 4.3.3.2  Non-Object Aliases
                  --  4.  Alternatively, if the name denotes a physical type
                  --      [...]
                  -- GHDL: this is not possible, since a physical type is
                  -- anonymous (LRM93 is buggy on this point).
                  if Get_Kind (Def) = Iir_Kind_Physical_Type_Definition then
                     raise Internal_Error;
                  end if;

                  --  LRM93 4.3.3.2  Non-Object Aliases
                  --  5.  Finally, if the name denotes a type, then implicit
                  --      alias declarations for each predefined operator
                  --      for the type immediatly follow the explicit alias
                  --      declaration for the type, and if present, any
                  --      implicit alias declarations for literals or units
                  --      of the type.
                  --      Each implicit alias has a signature that matches the
                  --      parameter and result type profule of the implicit
                  --      operator being aliased.
                  El := Get_Chain (N_Entity);
                  while El /= Null_Iir loop
                     case Get_Kind (El) is
                        when Iir_Kind_Implicit_Function_Declaration
                          | Iir_Kind_Implicit_Procedure_Declaration =>
                           exit when Get_Type_Reference (El) /= N_Entity;
                        when others =>
                           exit;
                     end case;
                     Add_Implicit_Alias (El);
                     El := Get_Chain (El);
                  end loop;
               end;
            when Iir_Kinds_Object_Declaration =>
               Error_Msg_Sem
                 ("non-object alias cannot denotes an object", Alias);
               --  Do not return and add the name to avoid an error storm.
            when Iir_Kind_Subtype_Declaration =>
               null;
            when others =>
               Error_Kind ("sem_non_object_alias_declaration", N_Entity);
         end case;
         Set_Name (Alias, N_Entity);
         Add_Name (N_Entity, Get_Identifier (Alias), False);
         Set_Visible_Flag (Alias, True);
      end if;
   end Sem_Non_Object_Alias_Declaration;

   procedure Sem_Group_Declaration (Group : Iir_Group_Declaration)
   is
      use Tokens;

      Constituent_List : Iir_Group_Constituent_List;
      Template : Iir_Group_Template_Declaration;
      Class, Prev_Class : Token_Type;
      El : Iir;
      El_Name : Iir;
      El_Entity : Iir_Entity_Class;
   begin
      Sem_Scopes.Add_Name (Group);
      Template := Find_Declaration (Get_Group_Template_Name (Group),
                                    Decl_Group_Template);
      if Template = Null_Iir then
         return;
      end if;
      Set_Group_Template_Name (Group, Template);
      Constituent_List := Get_Group_Constituent_List (Group);
      El_Entity := Get_Entity_Class_Entry_Chain (Template);
      Prev_Class := Tok_Eof;
      for I in Natural loop
         El := Get_Nth_Element (Constituent_List, I);
         exit when El = Null_Iir;

         if El_Entity = Null_Iir then
            Error_Msg_Sem
              ("too many elements in group constituent list", Group);
            exit;
         end if;

         Class := Get_Entity_Kind (El_Entity);
         if Class = Tok_Box then
            --  LRM93 4.6
            --  An entity class entry that includes a box (<>) allows zero
            --  or more group constituents to appear in this position in the
            --  corresponding group declaration.
            Class := Prev_Class;
         else
            Prev_Class := Class;
            El_Entity := Get_Chain (El_Entity);
         end if;

         Sem_Name (El);
         El_Name := Get_Named_Entity (El);
         if El_Name /= Error_Mark then
            --  LRM93 4.7
            --  It is an error if the class of any group constituent in the
            --  group constituent list is not the same as the class specified
            --  by the corresponding entity class entry in the entity class
            --  entry list of the group template.
            if Get_Entity_Class_Kind (El_Name) /= Class then
               Error_Msg_Sem
                 ("constituent not of class `" & Tokens.Image (Class) & ''',
                  El);
            end if;
         end if;
      end loop;

      --  End of entity_class list reached or zero or more constituent allowed.
      if not (El_Entity = Null_Iir
              or else Get_Entity_Kind (El_Entity) = Tok_Box)
      then
         Error_Msg_Sem
           ("not enough elements in group constituent list", Group);
      end if;
      Set_Visible_Flag (Group, True);
   end Sem_Group_Declaration;

   --  Semantize every declaration of DECLS_PARENT.
   --  STMTS is the concurrent statement list associated with DECLS_PARENT
   --  if any, or null_iir.  This is used for specification.
   procedure Sem_Declaration_Chain (Parent : Iir)
   is
      Decl: Iir;
      Last_Decl : Iir;
   begin
      --  Due to implicit declarations, the list can grow during sem.
      Decl := Get_Declaration_Chain (Parent);
      Last_Decl := Null_Iir;
      loop
         << Again >> exit when Decl = Null_Iir;
         case Get_Kind (Decl) is
            when Iir_Kind_Type_Declaration =>
               Sem_Type_Declaration (Decl);
            when Iir_Kind_Anonymous_Type_Declaration =>
               Sem_Type_Declaration (Decl);
            when Iir_Kind_Subtype_Declaration =>
               Sem_Subtype_Declaration (Decl);
            when Iir_Kind_Signal_Declaration =>
               Sem_Object_Declaration (Decl, Parent);
            when Iir_Kind_Constant_Declaration =>
               Sem_Object_Declaration (Decl, Parent);
            when Iir_Kind_Variable_Declaration =>
               Sem_Object_Declaration (Decl, Parent);
            when Iir_Kind_Attribute_Declaration =>
               Sem_Attribute_Declaration (Decl);
            when Iir_Kind_Attribute_Specification =>
               Sem_Attribute_Specification (Decl, Parent);
            when Iir_Kind_Component_Declaration =>
               Sem_Component_Declaration (Decl);
            when Iir_Kind_Function_Declaration
              | Iir_Kind_Procedure_Declaration =>
               declare
                  Res : Iir;
               begin
                  Res := Sem_Subprogram_Declaration (Decl);
                  if Res /= Decl then
                     --  Replace DECL with RES.
                     if Last_Decl = Null_Iir then
                        Set_Declaration_Chain (Parent, Res);
                     else
                        Set_Chain (Last_Decl, Res);
                     end if;
                     Decl := Res;
                     goto Again;
                  end if;
               end;
            when Iir_Kind_Function_Body
              | Iir_Kind_Procedure_Body =>
               Sem_Subprogram_Body (Decl);
            when Iir_Kind_Implicit_Function_Declaration
              | Iir_Kind_Implicit_Procedure_Declaration =>
               Sem_Scopes.Add_Name (Decl);
               Name_Visible (Decl);
            when Iir_Kind_Object_Alias_Declaration =>
               Sem_Object_Alias_Declaration (Decl);
            when Iir_Kind_Non_Object_Alias_Declaration =>
               Last_Decl := Decl;
               Decl := Get_Chain (Decl);
               Sem_Non_Object_Alias_Declaration (Last_Decl);
               goto Again;
            when Iir_Kind_File_Declaration =>
               Sem_File_Declaration (Decl);
            when Iir_Kind_Use_Clause =>
               Sem_Use_Clause (Decl);
            when Iir_Kind_Configuration_Specification =>
               null;
            when Iir_Kind_Disconnection_Specification =>
               Sem_Disconnect_Specification (Decl);
            when Iir_Kind_Group_Template_Declaration =>
               Sem_Group_Template_Declaration (Decl);
            when Iir_Kind_Group_Declaration =>
               Sem_Group_Declaration (Decl);
            when Iir_Kinds_Signal_Attribute =>
               --  Added by sem, so nothing to do.
               null;
            when Iir_Kind_Protected_Type_Body =>
               Sem_Protected_Type_Body (Decl);
            when others =>
               Error_Kind ("sem_declaration_chain", Decl);
         end case;
         Last_Decl := Decl;
         Decl := Get_Chain (Decl);
      end  loop;

      case Get_Kind (Parent) is
         when Iir_Kind_Package_Declaration
           | Iir_Kind_Protected_Type_Declaration
           | Iir_Kind_Entity_Declaration =>
            --  Bodies can be declared in package body or architecture.
            --  FIXME: entity ??
            null;
         when others =>
            Check_Full_Declaration (Parent, Parent);
      end case;
   end Sem_Declaration_Chain;


   procedure Check_Full_Declaration (Decls_Parent : Iir; Decl: Iir)
   is
      El: Iir;
   begin
      --  LRM 3.5 Protected types.
      --  Each protected type declaration appearing immediatly within a given
      --  declaration region must have exactly one corresponding protected type
      --  body appearing immediatly within the same declarative region and
      --  textually subsequent to the protected type declaration.

      --  LRM 3.3.1 Incomplete type declarations
      --  For each incomplete type declaration, there must be a corresponding
      --  full type declaration with the same identifier.  This full type
      --  declaration must occur later and immediatly within the same
      --  declarative part as the incomplete type declaration to which it
      --  correspinds.

      --  LRM 4.3.1.1 Constant declarations
      --  If the assignment symbol ":=" followed by an expression is not
      --  present in a constant declaration, then the declaration declares a
      --  deferred constant.  Such a constant declaration must appear in a
      --  package declaration.  The corresponding full constant declaration,
      --  which defines the value of the constant, must appear in the body of
      --  the package (see 2.6).

      --  LRM 2.2 Subprogram bodies
      --  If both a declaration and a body are given, [...].  Furthermore,
      --  both the declaration and the body must occur immediatly within the
      --  same declaration region.
      El := Get_Declaration_Chain (Decls_Parent);
      while El /= Null_Iir loop
         case Get_Kind (El) is
            when Iir_Kind_Constant_Declaration =>
               if Get_Deferred_Declaration_Flag (El) then
                  if Get_Deferred_Declaration (El) = Null_Iir then
                     Error_Msg_Sem ("missing value for constant declared at "
                                    & Disp_Location (El), Decl);
                  else
                     --  Remove from visibility the full declaration of the
                     --  constant.
                     Set_Deferred_Declaration (El, Null_Iir);
                  end if;
               end if;
            when Iir_Kind_Function_Declaration
              | Iir_Kind_Procedure_Declaration =>
               if Get_Subprogram_Body (El) = Null_Iir then
                  Error_Msg_Sem ("missing body for " & Disp_Node (El)
                                 & " declared at "
                                 & Disp_Location (El), Decl);
               end if;
            when Iir_Kind_Type_Declaration =>
               declare
                  Def : Iir;
               begin
                  Def := Get_Type (El);
                  if Get_Kind (Def) = Iir_Kind_Incomplete_Type_Definition
                    and then Get_Type_Declarator (Def) = El
                  then
                     Error_Msg_Sem ("missing full type declaration for "
                                    & Disp_Node (El), El);
                  elsif Get_Kind (Def) = Iir_Kind_Protected_Type_Declaration
                    and then Get_Protected_Type_Body (Def) = Null_Iir
                  then
                     Error_Msg_Sem ("missing protected type body for "
                                    & Disp_Node (El), El);
                  end if;
               end;
            when others =>
               null;
         end case;
         El := Get_Chain (El);
      end loop;
   end Check_Full_Declaration;

   procedure Sem_Iterator (Iterator : Iir_Iterator_Declaration;
                           Staticness : Iir_Staticness)
   is
      It_Type: Iir;
      A_Range: Iir;
      Range_Type : Iir;
   begin
      It_Type := Get_Type (Iterator);
      A_Range := Sem_Discrete_Range_Expression (It_Type, Null_Iir);
      if A_Range = Null_Iir then
         Set_Type (Iterator, Null_Iir);
         return;
      end if;
      if Get_Kind (A_Range) in Iir_Kinds_Type_And_Subtype_Definition then
         Range_Type := A_Range;
      else
         Range_Type := Get_Type (A_Range);
      end if;
      case Get_Kind (Range_Type) is
         when Iir_Kinds_Discrete_Type_Definition =>
            null;
         when others =>
            Error_Msg_Sem ("iterator is not of discrete type", A_Range);
            Set_Type (Iterator, Null_Iir);
            return;
      end case;

      Set_Type (Iterator, Range_To_Subtype_Definition (A_Range));
      Set_Base_Name (Iterator, Iterator);
      Set_Expr_Staticness (Iterator, Staticness);
   end Sem_Iterator;


   --  Semantize all arguments of ASSOC_CHAIN
   --  Return TRUE if no error.
   function Sem_Actual_Of_Association_Chain (Assoc_Chain : Iir)
     return Boolean
   is
      Has_Named : Boolean;
      Ok : Boolean;
      Assoc : Iir;
      Res : Iir;
      Formal : Iir;
   begin
      -- Semantize all arguments
      -- OK is false if there is an error during semantic of one of the
      -- argument, but continue semantisation.
      Has_Named := False;
      Ok := True;
      Assoc := Assoc_Chain;
      while Assoc /= Null_Iir loop
         Formal := Get_Formal (Assoc);
         if Formal /= Null_Iir then
            Has_Named := True;
            --  FIXME: check FORMAL is well composed.
         elsif Has_Named then
            Error_Msg_Sem ("positional argument after named argument", Assoc);
            Ok := False;
         end if;
         if Get_Kind (Assoc) /= Iir_Kind_Association_Element_Open then
            Res := Sem_Expression_Ov (Get_Actual (Assoc), Null_Iir);
            if Res = Null_Iir then
               Ok := False;
            else
               Set_Actual (Assoc, Res);
            end if;
         end if;
         Assoc := Get_Chain (Assoc);
      end loop;
      return Ok;
   end Sem_Actual_Of_Association_Chain;

   --  Check class of EXPR may be associated with class of INTERFACE.
   procedure Check_Association_Class
     (Interface: Iir;
      Expr : Iir;
      Assoc : Iir_Association_Element_By_Expression)
   is
      Prefix : Iir;
   begin
      case Get_Kind (Interface) is
         when Iir_Kind_Signal_Interface_Declaration =>
            Prefix := Iirs_Utils.Get_Object_Prefix (Expr);
            --  LRM93 2.1.1
            --  In a subprogram call, the actual designator associated with a
            --  formal parameter of class signal must be a signal.
            case Get_Kind (Prefix) is
               when Iir_Kind_Signal_Interface_Declaration
                 | Iir_Kind_Signal_Declaration
                 | Iir_Kind_Guard_Signal_Declaration =>
                  --  LRM93 2.1.1.2
                  --  If an actual signal is associated with a signal parameter
                  --  of any mode, the actual must be denoted by a static
                  --  signal name.
                  if Get_Name_Staticness (Expr) < Globally then
                     Error_Msg_Sem ("actual signal must be static", Expr);
                  end if;
               when Iir_Kinds_Signal_Attribute =>
                  null;
               when others =>
                  Error_Msg_Sem
                    ("signal parameter requires a signal expression",
                     Assoc);
            end case;
         when Iir_Kind_Variable_Interface_Declaration =>
            Prefix := Iirs_Utils.Get_Object_Prefix (Expr);
            --  LRM93 2.1.1
            --  The actual designator associated with a formal of class
            --  variable must be a variable.
            case Get_Kind (Prefix) is
               when Iir_Kind_Variable_Interface_Declaration
                 | Iir_Kind_Variable_Declaration =>
                  null;
               when Iir_Kind_File_Interface_Declaration
                 | Iir_Kind_File_Declaration =>
                  --  LRM87 4.3.1.4
                  --  Such an object is a member of the variable class of
                  --  objects;
                  if Flags.Vhdl_Std >= Vhdl_93 then
                     Error_Msg_Sem
                       ("in vhdl 93, variable parameter cannot be a file",
                        Assoc);
                  end if;
               when others =>
                  Error_Msg_Sem ("variable parameter must be a variable",
                                 Assoc);
            end case;
         when Iir_Kind_File_Interface_Declaration =>
            Prefix := Iirs_Utils.Get_Object_Prefix (Expr);
            --  LRM93 2.1.1
            --  The actual designator associated with a formal of class file
            --  must be a file.
            case Get_Kind (Prefix) is
               when Iir_Kind_File_Interface_Declaration
                 | Iir_Kind_File_Declaration =>
                  null;
               when Iir_Kind_Variable_Declaration
                 | Iir_Kind_Variable_Interface_Declaration =>
                  if Flags.Vhdl_Std >= Vhdl_93 then
                     Error_Msg_Sem
                       ("in vhdl93, file parameter must be a file", Assoc);
                  end if;
               when others =>
                  Error_Msg_Sem ("file parameter must be a file", Assoc);
            end case;
         when Iir_Kind_Constant_Interface_Declaration =>
            --  LRM93 2.1.1
            --  The actual designator associated with a formal of class
            --  constant must be an expression.
            null;
         when others =>
            Error_Kind ("check_association_class", Interface);
      end case;
   end Check_Association_Class;

   procedure Check_Unassociated
     (Interface : Iir; Assoc : Iir; Is_Port : Boolean; Loc : Iir)
   is
      procedure Error (Str : String) is
      begin
         if Assoc = Null_Iir then
            Error_Msg_Sem (Str, Loc);
         else
            Error_Msg_Sem (Str, Assoc);
         end if;
      end Error;
   begin
      if Get_Default_Value (Interface) /= Null_Iir then
         return;
      end if;
      if not Is_Port then
         if Assoc = Null_Iir then
            Error_Msg_Sem ("no parameter for " & Disp_Node (Interface), Loc);
         else
            Error_Msg_Sem ("open association not allowed", Assoc);
         end if;
         return;
      end if;

      -- LRM93 1.1.1.2
      -- A port of mode IN may be unconnected or unassociated only if its
      -- declaration includes a default expression.
      case Get_Mode (Interface) is
         when Iir_In_Mode =>
            Error (Disp_Node (Interface) & " of mode IN must be connected");
         when Iir_Out_Mode
           | Iir_Linkage_Mode
           | Iir_Inout_Mode
           | Iir_Buffer_Mode =>
            if Is_Unconstrained_Type_Definition (Get_Type (Interface)) then
               Error ("unconstrained " & Disp_Node (Interface)
                      & " must be connected");
            end if;
         when Iir_Unknown_Mode =>
            raise Internal_Error;
      end case;
   end Check_Unassociated;

   --  EXPR is a formal or actual expression.
   --  Extract conversion function CONV from EXPR, if:
   --  * argument of the function is of type ARG_TYPE.
   --  * return type of the function is RES_TYPE if RES_TYPE /= Null_Iir
   --    or any type if RES_TYPE = Null_Iir.
   procedure Sem_Conversion
     (Expr : in out Iir; Conv : out Iir)
   is
      Assoc : Iir;
   begin
      Conv := Null_Iir;
      case Get_Kind (Expr) is
         when Iir_Kind_Parenthesis_Name =>
            raise Internal_Error;
         when Iir_Kind_Function_Call =>
            null;
         when others =>
            return;
      end case;
      Conv := Get_Implementation (Expr);
      Assoc := Get_Parameter_Association_Chain (Expr);
      Expr := Get_Actual (Assoc);
      Free_Iir (Assoc);
   end Sem_Conversion;

   --  Finish to semantize association ASSOC associated with INTERFACE.
   --  FORMAL_TYPES is a list of possibles formal types.
   --  FCONV_TYPES is a list of possibles formal conversion types.
   procedure Sem_Association
     (Assoc : Iir;
      Interface : Iir;
      Formal_Types : Iir;
      Fconv_Types : Iir;
      Is_Port : Boolean;
      Loc : Iir)
   is
      Formal, Actual : Iir;
      Expr : Iir;
      Out_Conv, In_Conv : Iir;
      Formal_Type : Iir;
   begin
      Set_Associated_Formal (Assoc, Interface);

      In_Conv := Null_Iir;
      Out_Conv := Null_Iir;

      --  Semantize formal.
      Formal := Get_Formal (Assoc);
      if Formal = Null_Iir then
         Formal := Interface;
         Formal_Type := Get_Type (Interface);
      else
         if Fconv_Types /= Null_Iir then
            Sem_Conversion (Formal, Out_Conv);
            Set_Out_Conversion (Assoc, Out_Conv);
         end if;
         Formal_Type := Formal_Types;
      end if;

      if Get_Kind (Assoc) /= Iir_Kind_Association_Element_Open then
         Actual := Get_Actual (Assoc);
      end if;

--       if Get_Kind (Assoc) = Iir_Kind_Association_Element_Open then
--          Formal_Type := Null_Iir;
--       else
--          Actual := Get_Actual (Assoc);
--          if Fconv_Types = Null_Iir then
--             Formal_Type := Formal_Types;
--          else
--             Formal_Type := Fconv_Types;
--          end if;
--          Formal_Type :=
--            Search_Compatible_Type (Formal_Type, Get_Type (Actual));

--          if Formal_Type = Null_Iir then
--             Error_Msg_Sem ("actual and formal types mismatch", Assoc);
--             return;
--          end if;
--       end if;
      Set_Formal (Assoc, Formal);

      --  Semantize actual.
      if Get_Kind (Assoc) = Iir_Kind_Association_Element_Open then
         Check_Unassociated (Interface, Assoc, Is_Port, Loc);
      else
         if Get_Kind (Interface) /= Iir_Kind_Constant_Interface_Declaration
         then
            Sem_Conversion (Actual, In_Conv);
            Set_In_Conversion (Assoc, In_Conv);
         end if;
         Expr := Sem_Expression (Actual, Formal_Type);
         if Expr /= Null_Iir then
            Expr := Eval_Expr_If_Static (Expr);
            Set_Actual (Assoc, Expr);
            if In_Conv = Null_Iir and then Out_Conv = Null_Iir then
               if not Check_Implicit_Conversion (Get_Type (Formal), Expr) then
                  Error_Msg_Sem ("actual length does not match formal length",
                                 Assoc);
               end if;
            end if;
            if not Is_Port then
               Check_Association_Class (Interface, Expr, Assoc);
            end if;
         end if;
      end if;
   end Sem_Association;

   --  Return a type or a list of types of all possible conversion
   --  functions if the parenthesis name formal of ASSOC is a conversion
   --  function for INTERFACE.
   --  Return NULL_IIR if no match.
   --  If FINISH is true, then performs substitutions (and convert parenthesis
   --   name to function_call)
   function Conversion_Match (Assoc : Iir; Interface : Iir; Finish : Boolean)
     return Iir
   is
      Assoc_El : Iir;
      El : Iir;
      Act : Iir;
      Res : Iir;
      Call : Iir;
      Inter : Iir;
      Name : Iir;

      function Extract_Return_Type (Func : Iir) return Iir is
      begin
         case Get_Kind (Func) is
            when Iir_Kinds_Function_Declaration =>
               return Get_Return_Type (Func);
            when others =>
               return Null_Iir;
         end case;
      end Extract_Return_Type;
   begin
      Name := Get_Formal (Assoc);

      --  LRM93 4.3.2.2
      --  [...] whose single parameter of the function [...]
      Assoc_El := Get_Association_Chain (Name);
      if not Is_Chain_Length_One (Assoc_El) then
         return Null_Iir;
      end if;
      if Get_Kind (Assoc_El) /= Iir_Kind_Association_Element_By_Expression then
         return Null_Iir;
      end if;

      --  LRM93 4.3.2.2
      --  [...] is the formal designator itself
      Act := Get_Actual (Assoc_El);
      --  FIXME: the designator can be a slice, a selected or an indexed name.
      if Get_Kind (Act) /= Iir_Kind_Simple_Name
        or else Get_Identifier (Act) /= Get_Identifier (Interface)
      then
         return Null_Iir;
      end if;

      --  Get declarations for the prefix, extract return type.
      Inter := Get_Prefix (Name);
      Sem_Name (Inter);
      Inter := Get_Named_Entity (Inter);
      if Inter = Null_Iir then
         return Null_Iir;
      end if;
      if Is_Overload_List (Inter) then
         declare
            List : Iir_List;
         begin
            List := Create_Iir_List;
            for I in Natural loop
               El := Get_Nth_Element (Get_Overload_List (Inter), I);
               exit when El = Null_Iir;
               El := Extract_Return_Type (El);
               if El /= Null_Iir then
                  Append_Element (List, El);
               end if;
            end loop;
            Res := Simplify_Overload_List (List);
         end;
      else
         Res := Extract_Return_Type (Inter);
      end if;
      if not Finish then
         if Is_Overload_List (Inter) then
            Free_Iir (Inter);
         end if;
         return Res;
      end if;

      Call := Create_Iir (Iir_Kind_Function_Call);
      Location_Copy (Call, Name);
      Set_Implementation (Call, Inter);
      Set_Base_Name (Call, Call);
      Set_Parameter_Association_Chain (Call, Assoc_El);
      Free_Iir (Act);
      Set_Actual (Assoc_El, Interface); --FIXME bug here?
      Set_Type (Call, Res);
      Set_Formal (Assoc, Call);
      return Res;
   end Conversion_Match;

   --  Return a type or a list of types for a formal expression FORMAL
   --   corresponding to INTERFACE.  Possible cases are:
   --  * FORMAL is the simple name with the same identifier as INTERFACE,
   --    FORMAL_TYPE is set to the type of INTERFACE and CONV_TYPE is set
   --    to NULL_IIR.
   --  * FORMAL is a selected, indexed or slice name whose extreme prefix is
   --    a simple name with the same identifier as INTERFACE, FORMAL_TYPE
   --    is set to the type of the name, and CONV_TYPE is set to NULL_IIR.
   --  * FORMAL is a function call, whose only argument is an
   --    association_element_by_expression, whose actual is a name
   --    whose prefix is the same identifier as INTERFACE (note, since FORMAL
   --    is not semantized, this is parenthesis name), CONV_TYPE is set to
   --    the type or list of type of return type of conversion functions and
   --    FORMAL_TYPE is set to the type of the name.
   --  * otherwise, FORMAL cannot match INTERFACE and both FORMAL_TYPE and
   --    CONV_TYPE are set to NULL_IIR.
   --  If FINISH is true, the simple name is replaced by INTERFACE.
   procedure Sem_Formal (Assoc : Iir;
                         Interface : Iir;
                         Formal_Type : out Iir;
                         Conv_Type : out Iir;
                         Finish : Boolean)
   is
      Formal : Iir;
      Prefix : Iir;
      Res : Iir;
      Prefix_Type : Iir;
      Id : constant Name_Id := Get_Identifier (Interface);

      --  Set prefix and prefix_type.
      function Sem_Formal_Name (Name : Iir) return Boolean
      is
         Base_Type : Iir;
         Rec_El : Iir;
      begin
         case Get_Kind (Name) is
            when Iir_Kind_Simple_Name =>
               if Get_Identifier (Name) = Id then
                  Prefix := Name;
                  Prefix_Type := Get_Type (Interface);
                  return True;
               else
                  return False;
               end if;
            when Iir_Kind_Selected_Name =>
               if not Sem_Formal_Name (Get_Prefix (Name)) then
                  return False;
               end if;
               Base_Type := Get_Base_Type (Prefix_Type);
               if Get_Kind (Base_Type) /= Iir_Kind_Record_Type_Definition then
                  return False;
               end if;
               Rec_El := Find_Name_In_Chain
                 (Get_Element_Declaration_Chain (Base_Type),
                  Get_Suffix_Identifier (Name));
               if Rec_El = Null_Iir then
                  return False;
               end if;
               Prefix_Type := Get_Type (Rec_El);
               return True;
            when Iir_Kind_Parenthesis_Name =>
               --  More difficult: slice or indexed array.
               if not Sem_Formal_Name (Get_Prefix (Name)) then
                  return False;
               end if;
               Base_Type := Get_Base_Type (Prefix_Type);
               if Get_Kind (Base_Type) /= Iir_Kind_Array_Type_Definition then
                  return False;
               end if;
               declare
                  Chain : Iir;
                  Index_List : Iir_List;
                  Idx : Iir;
               begin
                  Chain := Get_Association_Chain (Name);
                  Index_List := Get_Index_Subtype_List (Base_Type);
                  --  Check for matching length.
                  if Get_Chain_Length (Chain) /= Get_Nbr_Elements (Index_List)
                  then
                     return False;
                  end if;
                  if Get_Kind (Chain)
                    /= Iir_Kind_Association_Element_By_Expression
                  then
                     return False;
                  end if;
                  Idx := Get_Actual (Chain);
                  if (not Is_Chain_Length_One (Chain))
                    or else (Get_Kind (Idx) /= Iir_Kind_Range_Expression
                             and then not Is_Range_Attribute_Name (Idx))
                  --  FIXME: what about subtype !
                  then
                     --  Indexed name.
                     Prefix_Type := Get_Element_Subtype (Base_Type);
                     return True;
                  end if;
                  --  Slice.
                  return True;
               end;
            when others =>
               Error_Kind ("sem_formal_name)", Name);
         end case;
      end Sem_Formal_Name;
   begin
      Formal_Type := Null_Iir;
      Conv_Type := Null_Iir;
      Formal := Get_Formal (Assoc);
      case Get_Kind (Formal) is
         when Iir_Kind_Simple_Name =>
            if Get_Identifier (Formal) = Id then
               if Finish then
                  Free_Iir (Formal);
                  Set_Formal (Assoc, Interface);
               end if;
               Formal_Type := Get_Type (Interface);
               Set_Whole_Association_Flag (Assoc, True);
            end if;
            return;
         when Iir_Kind_Selected_Name
           | Iir_Kind_Slice_Name =>
            null;
         when Iir_Kind_Parenthesis_Name =>
            if Get_Kind (Interface) /= Iir_Kind_Constant_Interface_Declaration
            then
               Res := Conversion_Match (Assoc, Interface, Finish);
               if Res /= Null_Iir then
                  Set_Whole_Association_Flag (Assoc, True);
                  Conv_Type := Res;
                  Formal_Type := Get_Type (Interface);
                  return;
               end if;
            end if;
            --  Maybe the prefix is a subelement of the interface.
            --  Will be checked just below.
         when others =>
            --  Should have been caught by sem_association_list.
            Error_Kind ("sem_formal", Formal);
      end case;
      --  Check for a sub-element.
      Set_Whole_Association_Flag (Assoc, False);
      if Sem_Formal_Name (Formal) then
         if Finish then
            Set_Named_Entity (Prefix, Interface);
            Formal := Sem_Expression (Formal, Prefix_Type);
            if Formal /= Null_Iir then
               Set_Formal (Assoc, Formal);
               Formal_Type := Prefix_Type;
            end if;
         else
            Formal_Type := Prefix_Type;
         end if;
      else
         Formal_Type := Null_Iir;
      end if;
   end Sem_Formal;

   --  Handle indexed name
   procedure Add_Individual_Assoc_Indexed_Name
     (Sub_Assoc : in out Iir; Formal : Iir)
   is
      Choice : Iir;
      Last_Choice : Iir;
      Index_List : Iir_List;
      Index : Iir;
      Nbr : Natural;
   begin
      --  Find element.
      Index_List := Get_Index_List (Formal);
      Nbr := Get_Nbr_Elements (Index_List);
      for I in 0 .. Nbr - 1 loop
         Index := Get_Nth_Element (Index_List, I);

         --  Evaluate index.
         Index := Eval_Expr (Index);
         Replace_Nth_Element (Index_List, I, Index);

         --  Find index in choice list.
         Last_Choice := Null_Iir;
         Choice := Get_Individual_Association_Chain (Sub_Assoc);
         while Choice /= Null_Iir loop
            case Get_Kind (Choice) is
               when Iir_Kind_Choice_By_Expression =>
                  if Eval_Pos (Get_Expression (Choice)) = Eval_Pos (Index) then
                     goto Found;
                  end if;
               when Iir_Kind_Choice_By_Range =>
                  if Eval_Int_In_Range (Eval_Pos (Index),
                                        Get_Expression (Choice))
                  then
                     --  FIXME: overlap.
                     raise Internal_Error;
                  end if;
               when others =>
                  Error_Kind ("add_individual_assoc_index_name", Choice);
            end case;
            Last_Choice := Choice;
            Choice := Get_Chain (Choice);
         end loop;

         --  If not found, append it.
         Choice := Create_Iir (Iir_Kind_Choice_By_Expression);
         Set_Expression (Choice, Index);
         Location_Copy (Choice, Formal);
         if Last_Choice = Null_Iir then
            Set_Individual_Association_Chain (Sub_Assoc, Choice);
         else
            Set_Chain (Last_Choice, Choice);
         end if;

         << Found >> null;

         if I < Nbr - 1 then
            Sub_Assoc := Get_Associated (Choice);
            if Sub_Assoc = Null_Iir then
               Sub_Assoc := Create_Iir
                 (Iir_Kind_Association_Element_By_Individual);
               Location_Copy (Sub_Assoc, Index);
               Set_Associated (Choice, Sub_Assoc);
            end if;
         else
            Sub_Assoc := Choice;
         end if;
      end loop;
   end Add_Individual_Assoc_Indexed_Name;

   procedure Add_Individual_Assoc_Slice_Name
     (Sub_Assoc : in out Iir; Formal : Iir)
   is
      Choice : Iir;
      Index : Iir;
   begin
      --  FIXME: handle cases such as param(5 to 6)(5)

      --  Find element.
      Index := Get_Suffix (Formal);

      --  Evaluate index.
      Index := Eval_Expr (Index);
      Set_Suffix (Formal, Index);

      Choice := Create_Iir (Iir_Kind_Choice_By_Range);
      Location_Copy (Choice, Formal);
      Set_Expression (Choice, Index);
      Set_Chain (Choice, Get_Individual_Association_Chain (Sub_Assoc));
      Set_Individual_Association_Chain (Sub_Assoc, Choice);

      Sub_Assoc := Choice;
   end Add_Individual_Assoc_Slice_Name;

   procedure Add_Individual_Assoc_Selected_Name
     (Sub_Assoc : in out Iir; Formal : Iir)
   is
      Choice : Iir;
   begin
      Choice := Create_Iir (Iir_Kind_Choice_By_Name);
      Location_Copy (Choice, Formal);
      Set_Name (Choice, Get_Selected_Element (Formal));
      Set_Chain (Choice, Get_Individual_Association_Chain (Sub_Assoc));
      Set_Individual_Association_Chain (Sub_Assoc, Choice);

      Sub_Assoc := Choice;
   end Add_Individual_Assoc_Selected_Name;

   procedure Add_Individual_Association_1 (Iassoc : in out Iir; Formal : Iir)
   is
      Sub : Iir;
   begin
      case Get_Kind (Formal) is
         when Iir_Kind_Indexed_Name
           | Iir_Kind_Slice_Name
           | Iir_Kind_Selected_Element =>
            Add_Individual_Association_1 (Iassoc, Get_Prefix (Formal));
         when Iir_Kinds_Interface_Declaration =>
            return;
         when others =>
            Error_Kind ("add_individual_association_1", Formal);
      end case;

      case Get_Kind (Iassoc) is
         when Iir_Kind_Association_Element_By_Individual =>
            null;
         when Iir_Kind_Choice_By_Expression =>
            Sub := Get_Associated (Iassoc);
            if Sub = Null_Iir then
               Sub := Create_Iir (Iir_Kind_Association_Element_By_Individual);
               Location_Copy (Sub, Formal);
               Set_Associated_Formal (Sub, Formal);
               Set_Formal (Sub, Iassoc);
               Iassoc := Sub;
            else
               case Get_Kind (Sub) is
                  when Iir_Kind_Association_Element_By_Individual =>
                     Iassoc := Sub;
                  when others =>
                     Error_Msg_Sem
                       ("individual association of "
                        & Disp_Node (Get_Associated_Formal (Iassoc))
                        & " conflicts with that at " & Disp_Location (Sub),
                        Formal);
                     return;
               end case;
            end if;
         when others =>
            Error_Kind ("add_individual_association_1(2)", Iassoc);
      end case;

      case Get_Kind (Formal) is
         when Iir_Kind_Indexed_Name =>
            Add_Individual_Assoc_Indexed_Name (Iassoc, Formal);
         when Iir_Kind_Slice_Name =>
            Add_Individual_Assoc_Slice_Name (Iassoc, Formal);
         when Iir_Kind_Selected_Element =>
            Add_Individual_Assoc_Selected_Name (Iassoc, Formal);
         when others =>
            Error_Kind ("add_individual_association_1(3)", Formal);
      end case;
   end Add_Individual_Association_1;

   procedure Add_Individual_Association (Iassoc : Iir; Assoc : Iir)
   is
      Formal : Iir;
      Iass : Iir;
      Prev : Iir;
   begin
      Formal := Get_Formal (Assoc);
      Iass := Iassoc;
      Add_Individual_Association_1 (Iass, Formal);
      Prev := Get_Associated (Iass);
      if Prev /= Null_Iir then
         Error_Msg_Sem ("individual association of "
                        & Disp_Node (Get_Associated_Formal (Assoc))
                        & " conflicts with that at " & Disp_Location (Prev),
                        Assoc);
      else
         Set_Associated (Iass, Get_Actual (Assoc));
      end if;
   end Add_Individual_Association;

   procedure Finish_Individual_Assoc_Array_Subtype (Assoc : Iir; Atype : Iir)
   is
      Index_Tlist : Iir_List;
      Index_Type : Iir;
      Low, High : Iir;
   begin
      Index_Tlist := Get_Index_Subtype_List (Atype);
      for I in Natural loop
         Index_Type := Get_Nth_Element (Index_Tlist, I);
         exit when Index_Type = Null_Iir;
         Sem_Choices_Range
           (Get_Individual_Association_Chain (Assoc), Index_Type, True,
            Get_Location (Assoc), Low, High);
      end loop;
   end Finish_Individual_Assoc_Array_Subtype;

   procedure Finish_Individual_Assoc_Array
     (Actual : Iir; Assoc : Iir; Dim : Natural)
   is
      Actual_Type : Iir;
      Actual_Index : Iir;
      Base_Type : Iir;
      Base_Index : Iir;
      Low, High : Iir;
   begin
      Actual_Type := Get_Actual_Type (Actual);
      Actual_Index := Get_Nth_Element (Get_Index_Subtype_List (Actual_Type),
                                       Dim - 1);
      if Actual_Index /= Null_Iir then
         Base_Index := Actual_Index;
      else
         Base_Type := Get_Base_Type (Actual_Type);
         Base_Index := Get_Nth_Element (Get_Index_Subtype_List (Base_Type),
                                        Dim - 1);
      end if;
      Sem_Choices_Range
        (Get_Individual_Association_Chain (Assoc), Base_Index, True,
         Get_Location (Assoc), Low, High);
      if Actual_Index = Null_Iir then
         declare
            Index_Constraint : Iir;
            Index_Subtype_Constraint : Iir;
         begin
            --  Create an index subtype.
            case Get_Kind (Base_Index) is
               when Iir_Kind_Integer_Subtype_Definition =>
                  Actual_Index :=
                    Create_Iir (Iir_Kind_Integer_Subtype_Definition);
               when Iir_Kind_Enumeration_Type_Definition
                 | Iir_Kind_Enumeration_Subtype_Definition =>
                  Actual_Index :=
                    Create_Iir (Iir_Kind_Enumeration_Subtype_Definition);
               when others =>
                  Error_Kind ("finish_individual_assoc_array", Base_Index);
            end case;
            Location_Copy (Actual_Index, Actual);
            Set_Base_Type (Actual_Index, Get_Base_Type (Base_Index));
            Index_Constraint := Get_Range_Constraint (Base_Index);

            Index_Subtype_Constraint := Create_Iir (Iir_Kind_Range_Expression);
            Location_Copy (Index_Subtype_Constraint, Actual);
            Set_Range_Constraint (Actual_Index, Index_Subtype_Constraint);
            Set_Type_Staticness (Actual_Index, Locally);
            Set_Direction (Index_Subtype_Constraint,
                           Get_Direction (Index_Constraint));

            case Get_Direction (Index_Constraint) is
               when Iir_To =>
                  Set_Left_Limit (Index_Subtype_Constraint, Low);
                  Set_Right_Limit (Index_Subtype_Constraint, High);
               when Iir_Downto =>
                  Set_Left_Limit (Index_Subtype_Constraint, High);
                  Set_Right_Limit (Index_Subtype_Constraint, Low);
            end case;
            Set_Expr_Staticness (Index_Subtype_Constraint, Locally);
            Append_Element (Get_Index_Subtype_List (Actual_Type),
                            Actual_Index);
         end;
      else
         declare
            Act_High, Act_Low : Iir;
         begin
            Get_Low_High_Limit (Get_Range_Constraint (Actual_Type),
                                Act_Low, Act_High);
            if Eval_Pos (Act_Low) /= Eval_Pos (Low)
              or Eval_Pos (Act_High) /= Eval_Pos (High)
            then
               Error_Msg_Sem ("indexes of individual association mismatch",
                              Assoc);
            end if;
         end;
      end if;
   end Finish_Individual_Assoc_Array;

   procedure Finish_Individual_Assoc_Record (Assoc : Iir; Atype : Iir)
   is
      Base_Type : Iir_Record_Type_Definition;
      Matches : Iir_Array_Acc;
      Ch : Iir;
      Pos : Natural;
      Rec_El : Iir;
   begin
      Base_Type := Get_Base_Type (Atype);
      Matches := new Iir_Array
        (0 .. Natural (Get_Number_Element_Declaration (Base_Type)) - 1);
      Matches.all := (others => Null_Iir);
      Ch := Get_Individual_Association_Chain (Assoc);
      while Ch /= Null_Iir loop
         Rec_El := Get_Name (Ch);
         Pos := Natural (Get_Element_Position (Rec_El));
         if Matches (Pos) /= Null_Iir then
            Error_Msg_Sem ("individual " & Disp_Node (Rec_El)
                           & " already associated at "
                           & Disp_Location (Matches (Pos)), Ch);
         else
            Matches (Pos) := Ch;
         end if;
         Ch := Get_Chain (Ch);
      end loop;
      Rec_El := Get_Element_Declaration_Chain (Base_Type);
      for I in Matches'Range loop
         if Matches (I) = Null_Iir then
            Error_Msg_Sem (Disp_Node (Rec_El) & " not associated", Assoc);
         end if;
         Rec_El := Get_Chain (Rec_El);
      end loop;
      Set_Actual_Type (Assoc, Atype);
   end Finish_Individual_Assoc_Record;

   --  Called by sem_individual_association to finish the semantization of
   --  individual association ASSOC.
   procedure Finish_Individual_Association (Assoc : Iir)
   is
      Formal : Iir;
      Atype : Iir;
   begin
      --  Guard.
      if Assoc = Null_Iir then
         return;
      end if;

      Formal := Get_Associated_Formal (Assoc);
      Atype := Get_Type (Formal);

      case Get_Kind (Atype) is
         when Iir_Kind_Array_Subtype_Definition =>
            Finish_Individual_Assoc_Array_Subtype (Assoc, Atype);
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
            Set_Actual_Type
              (Assoc, Create_Array_Subtype (Atype, Get_Location (Assoc)));
            Finish_Individual_Assoc_Array (Assoc, Assoc, 1);
         when Iir_Kind_Record_Type_Definition
           | Iir_Kind_Record_Subtype_Definition =>
            Finish_Individual_Assoc_Record (Assoc, Atype);
         when others =>
            Error_Kind ("finish_individual_association", Atype);
      end case;
   end Finish_Individual_Association;

   --  Sem individual associations of ASSOCS:
   --  Add an Iir_Kind_Association_Element_By_Individual before each
   --  group of individual association for the same formal, and call
   --  Finish_Individual_Association with each of these added nodes.
   procedure Sem_Individual_Association (Assoc_Chain : in out Iir)
   is
      Assoc : Iir;
      Prev_Assoc : Iir;
      Iassoc : Iir_Association_Element_By_Individual;
      Cur_Iface : Iir;
   begin
      Iassoc := Null_Iir;
      Cur_Iface := Null_Iir;
      Prev_Assoc := Null_Iir;
      Assoc := Assoc_Chain;
      while Assoc /= Null_Iir loop
         if Get_Associated_Formal (Assoc) /= Cur_Iface then
            --  New formal name, sem the current assoc.
            Finish_Individual_Association (Iassoc);
            Cur_Iface := Get_Associated_Formal (Assoc);
            Iassoc := Null_Iir;
         end if;
         if Get_Whole_Association_Flag (Assoc) = False then
            --  New inidividual association.
            if Iassoc = Null_Iir then
               Iassoc :=
                 Create_Iir (Iir_Kind_Association_Element_By_Individual);
               Location_Copy (Iassoc, Assoc);
               Set_Associated_Formal (Iassoc, Cur_Iface);
               Set_Formal (Iassoc, Cur_Iface);
               --  Insert IASSOC.
               if Prev_Assoc = Null_Iir then
                  Assoc_Chain := Iassoc;
               else
                  Set_Chain (Prev_Assoc, Iassoc);
               end if;
               Set_Chain (Iassoc, Assoc);
            end if;
            Add_Individual_Association (Iassoc, Assoc);
         end if;
         Prev_Assoc := Assoc;
         Assoc := Get_Chain (Assoc);
      end loop;
      --  There is maybe a remaining iassoc.
      Finish_Individual_Association (Iassoc);
   end Sem_Individual_Association;

   procedure Sem_Association_Chain
     (Interface_Chain : Iir;
      Assoc_Chain: in out Iir;
      Finish: Boolean;
      Check_Match : Boolean;
      Is_Port : Boolean;
      Loc : Iir;
      Level : out Compatibility_Level)
   is
      Assoc: Iir;
      Interface: Iir;

      --  Associate ASSOC with interface INTERFACE
      --  This sets RES.
      procedure Try_Association (Formal_Types : Iir; Fconv_Type : Iir)
      is
         Actual: Iir;
         Actual_Type : Iir;
         Aconv_Type : Iir;
         Cassoc_Chain : Iir;
         Cassoc : Iir;
      begin
         if Get_Kind (Assoc) /= Iir_Kind_Association_Element_Open then
            Actual := Get_Actual (Assoc);
            Actual_Type := Get_Type (Actual);
            if Actual_Type = Null_Iir then
               Level := None;
               return;
            end if;
            Aconv_Type := Null_Iir;
            if Get_Kind (Actual) = Iir_Kind_Function_Call
              and then Get_Kind (Interface)
              /= Iir_Kind_Constant_Interface_Declaration
            then
               Cassoc_Chain := Get_Parameter_Association_Chain (Actual);
               if Is_Chain_Length_One (Cassoc_Chain) then
                  Cassoc := Cassoc_Chain;
                  if Get_Kind (Cassoc)
                    = Iir_Kind_Association_Element_By_Expression
                  then
                     Aconv_Type := Actual_Type;
                     Actual_Type := Get_Type (Get_Actual (Cassoc));
                  end if;
               end if;
            end if;

            --  4 cases:
            --    A  => B     type of A = type of B
            --  F(A) => B     type of B = type of F
            --    A  => G(B)  type of A = type of G
            --  F(A) => G(B)  type of B = type of F, type of A = type of G
            if Fconv_Type = Null_Iir and then Aconv_Type = Null_Iir then
               Level := Compatibility_Level'Min
                 (Level, Compatibility_Types (Formal_Types, Actual_Type));
            else
               if Fconv_Type /= Null_Iir then
                  Level := Compatibility_Level'Min
                    (Level, Compatibility_Types (Fconv_Type, Actual_Type));
               end if;
               if Aconv_Type /= Null_Iir then
                  Level := Compatibility_Level'Min
                    (Level, Compatibility_Types (Formal_Types, Aconv_Type));
               end if;
            end if;

            if Level = None then
               if Finish then
                  Not_Match (Actual, Interface);
               end if;
               return;
            end if;
         end if;
         if Finish then
            Sem_Association
              (Assoc, Interface, Formal_Types, Fconv_Type, Is_Port, Loc);
         end if;
      end Try_Association;

      type Param_Assoc_Type is (None, Individual, Whole);
      type Bool_Array is array (Natural range <>) of Param_Assoc_Type;
      Nbr_Arg: constant Natural := Get_Chain_Length (Interface_Chain);
      Arg_Matched: Bool_Array (0 .. Nbr_Arg - 1) := (others => None);
      Last_Individual : Iir;
      Formal_Types : Iir;
      Conv_Type : Iir;
      Has_Individual : Boolean;
      Pos : Integer;
   begin
      Level := Full;
      Has_Individual := False;

      -- Loop on every assoc element, try to match it.
      Interface := Interface_Chain;
      Last_Individual := Null_Iir;
      Pos := 0;

      Assoc := Assoc_Chain;
      while Assoc /= Null_Iir loop
         if Get_Formal (Assoc) = Null_Iir then
            -- Positional argument.
            if Pos < 0 then
               --  Positional after named argument.  Already caught.
               Level := None;
               return ;
            end if;
            -- Try to match actual of ASSOC with the interface.
            if Interface = Null_Iir then
               if Finish then
                  Error_Msg_Sem
                    ("too many arguments for " & Disp_Node (Loc), Assoc);
               end if;
               Level := None;
               return;
            end if;
            Try_Association (Get_Type (Interface), Null_Iir);
            if Level = None then
               return;
            end if;
            Arg_Matched (Pos) := Whole;
            Set_Whole_Association_Flag (Assoc, True);
            Interface := Get_Chain (Interface);
            Pos := Pos + 1;
         else
            -- Named association.
            Pos := -1;
            -- FIXME: directly search the formal if finish is true.
            -- Find the Interface.
            Interface := Interface_Chain;
            Pos := 0;
            while Interface /= Null_Iir loop
               -- Formal assoc is not necessarily a simple name, it may
               -- be a conversion function, or even an indexed or
               -- selected name.
               Sem_Formal (Assoc, Interface,
                           Formal_Types, Conv_Type, Finish);
               if Formal_Types /= Null_Iir then
                  Try_Association (Formal_Types, Conv_Type);
                  if Level = None then
                     return;
                  end if;
                  if Get_Whole_Association_Flag (Assoc) then
                     Last_Individual := Null_Iir;
                     if Arg_Matched (Pos) = None then
                        Arg_Matched (Pos) := Whole;
                     else
                        if Finish then
                           Error_Msg_Sem
                             (Disp_Node (Interface) & " already associated",
                              Assoc);
                           Level := None;
                           return;
                        end if;
                     end if;
                  else
                     Has_Individual := True;
                     if Arg_Matched (Pos) /= Whole then
                        if Finish
                          and then Arg_Matched (Pos) = Individual
                          and then Last_Individual /= Interface
                        then
                           Error_Msg_Sem
                             ("non consecutive individual association for "
                              & Disp_Node (Interface),
                              Assoc);
                           Level := None;
                           return;
                        end if;
                        Last_Individual := Interface;
                        Arg_Matched (Pos) := Individual;
                     else
                        if Finish then
                           Error_Msg_Sem
                             (Disp_Node (Interface) & " already associated",
                              Assoc);
                        end if;
                        Level := None;
                        return;
                     end if;
                  end if;
                  goto Found;
               end if;
               Interface := Get_Chain (Interface);
               Pos := Pos + 1;
            end loop;

            -- Not found.
            if Finish then
               Error_Msg_Sem ("no such interface name", Assoc);
            end if;
            Level := None;
            return;

            << Found >> null;
            Pos := -1;
         end if;
         Assoc := Get_Chain (Assoc);
      end loop;

      if Finish and then Has_Individual then
         Sem_Individual_Association (Assoc_Chain);
      end if;

      if not Check_Match or else Interface_Chain = Null_Iir then
         return;
      end if;

      --  LRM 8.6 Procedure Call Statement
      --  For each formal parameter of a procedure, a procedure call must
      --  specify exactly one corresponding actual parameter.
      --  This actual parameter is specified either explicitly, by an
      --  association element (other than the actual OPEN) in the association
      --  list, or in the absence of such an association element, by a default
      --  expression (see Section 4.3.3.2).

      --  LRM 7.3.3 Function Calls
      --  For each formal parameter of a function, a function call must
      --  specify exactly one corresponding actual parameter.
      --  This actual parameter is specified either explicitly, by an
      --  association element (other than the actual OPEN) in the association
      --  list, or n the absence of such an association element, by a default
      --  expression (see Section 4.3.3.2).
      Interface := Interface_Chain;
      Pos := 0;
      while Interface /= Null_Iir loop
         if Arg_Matched (Pos) = None
           and then Get_Default_Value (Interface) = Null_Iir
         then
            if Finish then
               Error_Msg_Sem ("no actual for " & Disp_Node (Interface), Loc);
            end if;
            Level := None;
            return;
         end if;
         Interface := Get_Chain (Interface);
         Pos := Pos + 1;
      end loop;
      return;
   end Sem_Association_Chain;

end Sem_Decls;
